import os
import re
import io
import hashlib
import pkg_resources
import json
import zlib
import base64
from copy import deepcopy
from io import BytesIO
from ruamel.yaml import YAML
from tempfile import mkdtemp
from shutil import rmtree
from urllib.parse import urlparse
from typing import MutableMapping, MutableSequence


def get_compressed(data_str, reset_position=None):
    """
    Converts character string "data_str" as "utf-8" into bytes ("utf-8"
    is default encoding for Python3 string). Encoded bytes are then being
    compressed with "zlib" and encoded again as "base64" (it uses only the
    characters A-Z, a-z, 0-9, +, /* so it can be transmitted over channels
    that do not preserve all 8-bits of data). At the end "base64" encoded
    copressed bytes are converted back to standard for Python3 "utf-8" string.
    In case we failed to encode "data_str" as string, assume it was an
    Object that can be dumped with json (useful for running tests). If we
    failed to dump it with json, assume that "data_str" was a stream, from
    where we read content either as "utf-8" or as bytes, depending on the mode
    that the file was opened with. In this case if "reset_position" is true,
    reset to the beginning of the file.
    """

    reset_position = True if reset_position is None else reset_position

    try:
        data_str_utf = data_str.encode("utf-8")
    except AttributeError:
        try:
            data_str_utf = json.dumps(data_str).encode("utf-8")
        except TypeError:
            if reset_position:
                data_str.seek(0)
            if not isinstance(data_str, io.TextIOBase):         # file was opened in a binary mode
                data_str_utf = data_str.read()
            else:                                               # file was opened in a text mode and need to be "utf-8" encoded
                data_str_utf = data_str.read().encode("utf-8")
    return base64.b64encode(
        zlib.compress(
            data_str_utf,
            level=9)
    ).decode("utf-8")


def get_uncompressed(data_str, parse_as_yaml=None):
    """
    Converts character string "data_str" as "utf-8" into bytes, then
    decodes it as "base64" and decompress with "zlib". The resulted
    "bytes" are converted again into standard for Python3 "utf-8"
    string. Raises zlib.error or binascii.Error if something went
    wrong. If "parse_as_yaml" is True, try to load uncompressed
    content with "load_yaml". The latter may raise ValueError or
    YAMLError if something went wrong
    """

    parse_as_yaml = False if parse_as_yaml is None else parse_as_yaml
    uncompressed =  zlib.decompress(
        base64.b64decode(
            data_str.encode("utf-8")
        )
    ).decode("utf-8")
    return load_yaml(uncompressed) if parse_as_yaml else uncompressed


def get_api_failure_reason(response):
    """
    Handy function to safely get a failure reason from
    the "detail" field of request.response object returned
    from our API
    """

    try:
        reason = response.json()["detail"]
    except (ValueError, KeyError):
        reason = "unknown reason"
    return reason


def remove_field_from_dict(data, key):
    """
    Returns data with all occurences of "key" removed.
    "data" should be a dictionary.
    """

    data_copy = deepcopy(data)

    def __clean(data, key):
        if isinstance(data, MutableMapping):
            if key in data:
                del data[key]
            for item in data:
                __clean(data[item], key)
        if isinstance(data, MutableSequence):
            for item in data:
                __clean(item, key)

    __clean(data_copy, key)

    return data_copy


def get_files(location, filename_pattern=None):
    """
    Recursively searches for files in a folder by regex pattern.
    Results for the files with the same basenames will be overwritten.
    """

    filename_pattern = ".*" if filename_pattern is None else filename_pattern
    files_dict = {}
    for root, dirs, files in os.walk(location):
        files_dict.update(
            {filename: os.path.join(root, filename) for filename in files if re.match(filename_pattern, filename)}
        )
    return files_dict


def get_dir(location, cwd=None, permissions=None, exist_ok=None):

    permissions = 0o0775 if permissions is None else permissions
    exist_ok = True if exist_ok is None else exist_ok
    cwd = os.getcwd() if cwd is None else cwd

    abs_location = get_absolute_path(location, cwd)
    try:
        os.makedirs(abs_location, mode=permissions)
    except os.error:
        if not exist_ok:
            raise
    return abs_location


def get_path_from_url(url):
    return urlparse(url).path


def get_absolute_path(p, cwd=None):
    """
    Get absolute path relative to cwd or current working directory
    """

    cwd = os.getcwd() if cwd is None else cwd

    return p if os.path.isabs(p) else os.path.normpath(os.path.join(cwd, p))


def get_rootname(location):
    return os.path.splitext(os.path.basename(location))[0]


def get_version():
    """
    Returns current version of the package if it's installed
    """

    pkg = pkg_resources.require("cwl_airflow")
    return pkg[0].version if pkg else "unknown version"


def get_md5_sum(location, block_size=2**20):
    """
    Calculates md5 sum of a file. If "location" cannot be
    opened, assumes that it was a file content in a form
    of "utf-8" string.
    """

    def __update_md5():
        while True:
            buf = input_stream.read(block_size)
            if not buf:
                break
            md5_sum.update(buf)

    md5_sum = hashlib.md5()
    try:
        url_path = get_path_from_url(location)                   # need to get rid of file:// if it was url
        with open(url_path , "rb") as input_stream:
            __update_md5()
    except (FileNotFoundError, OSError) as err:
        with BytesIO(location.encode("utf-8")) as input_stream:
            __update_md5() 

    return md5_sum.hexdigest()


def load_yaml(location):
    """
    Tries to load yaml document from file or string.

    If file cannot be loaded, assumes that location
    is a string and tries to load yaml from string.

    If string wasn't parsed and YAML didn't raise
    YAMLError, check ir the parsed result is the same
    as input. If yes, raise ValueError
    """

    yaml = YAML()
    yaml.preserve_quotes = True
    try:
        with open(location, "r") as input_stream:
            data = yaml.load(input_stream)
    except (FileNotFoundError, OSError):           # catch OSError raised when "filename too long"
        data = yaml.load(location)
    if data == location:
        raise ValueError
    return data


def dump_json(data, location):                    # TODO: consider substitute it with dump_yaml for consistency
    with open(location , "w") as output_stream:
        json.dump(data, output_stream, indent=4)


class CleanAirflowImport():
    """
    Replaces AIRFLOW_HOME and AIRFLOW_CONFIG from os.environ
    with temporary values. On exit either restores the previous values
    or removes them from os.environ, and cleans temp directory.
    Useful when importing modules from Airflow, that silently create
    airflow folder. Note, all the changes are made only within Python.
    __suppress_logging and __restore_logging are used to prevent Airflow
    from printing deprecation warnings
    """


    def __enter__(self):
        self.__suppress_logging()
        self.backup_airflow_home = os.environ.get("AIRFLOW_HOME")
        self.backup_airflow_config = os.environ.get("AIRFLOW_CONFIG")
        self.temp_airflow_home = mkdtemp()
        os.environ["AIRFLOW_HOME"] = self.temp_airflow_home
        os.environ["AIRFLOW_CONFIG"] = os.path.join(self.temp_airflow_home, "airflow.cfg")


    def __exit__(self, type, value, traceback):
        rmtree(self.temp_airflow_home)

        if self.backup_airflow_home is not None:
            os.environ["AIRFLOW_HOME"] = self.backup_airflow_home
        else:
            del os.environ["AIRFLOW_HOME"]

        if self.backup_airflow_config is not None:
            os.environ["AIRFLOW_CONFIG"] = self.backup_airflow_config
        else:
            del os.environ["AIRFLOW_CONFIG"]

        self.__restore_logging()


    def __suppress_logging(self):
        self.NULL_FDS = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        self.BACKUP_FDS = os.dup(1), os.dup(2)
        os.dup2(self.NULL_FDS[0], 1)
        os.dup2(self.NULL_FDS[1], 2)


    def __restore_logging(self):
        os.dup2(self.BACKUP_FDS[0], 1)
        os.dup2(self.BACKUP_FDS[1], 2)
        os.close(self.NULL_FDS[0])
        os.close(self.NULL_FDS[1])


#! /usr/bin/env python3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from cwl_airflow.utilities.cwl import (
    execute_workflow_step,
    get_containers,
    kill_containers,
    collect_reports
)
from cwl_airflow.utilities.report import post_status
from cwl_airflow.utilities.loggers import setup_cwl_logger


class CWLStepOperator(BaseOperator):


    @apply_defaults  # in case someone decided to overwrite default_args from the DAG
    def __init__(
        self,
        task_id,
        *args, **kwargs
    ):
        super().__init__(task_id=task_id, *args, **kwargs)


    def execute(self, context):
        """
        Creates job from collected reports of all finished tasks in a DAG.
        Then executes a workflow constructed from the workflow step. Writes
        report file location to X-Com.
        """

        setup_cwl_logger(context["ti"])
        post_status(context)

        self.job_data = collect_reports(context)         # we need it also in "on_kill"
        _, step_report = execute_workflow_step(
            workflow=context["dag"].workflow,
            task_id=self.task_id,
            job_data=self.job_data,
            cwl_args=context["dag"].default_args["cwl"]
        )

        return step_report


    def on_kill(self):
        """
        Function is called only if task is manually stopped, for example, from UI.
        First, we need to find all cidfiles that correspond to the current step.
        We can have more than one cidfile, if previous run of this step has failed.
        We search for cidfiles in the subfolder "task_id" of the "tmp_folder" read
        from "job_data". For all found cidfile we check if cointainer is still
        running and try to stop it. If container was not running, was not found or
        had been already successfully killed, we remove the correspondent cidfile.
        If container was running but we failed to kill it do not remove cidfile.
        """

        kill_containers(
            get_containers(self.job_data, self.task_id)
        )



import logging
import psutil
import shutil

from cwl_airflow.utilities.report import dag_on_success, dag_on_failure
from cwl_airflow.utilities.helpers import CleanAirflowImport

with CleanAirflowImport():
    from airflow import configuration
    from airflow.models import DAG, DagRun, TaskInstance
    from airflow.operators.python_operator import PythonOperator
    from airflow.utils.dates import days_ago
    from airflow.utils.db import provide_session
    from airflow.utils.state import State



logger = logging.getLogger(__name__)


TIMEOUT = configuration.conf.getint('core', 'KILLED_TASK_CLEANUP_TIME')


@provide_session
def clean_db(dr, session=None):
    logger.debug(f"""Clean DB for {dr.dag_id} - {dr.run_id}""")
    for ti in dr.get_task_instances():
        logger.debug(f"""process {ti.dag_id} - {ti.task_id} - {ti.execution_date}""")
        ti.clear_xcom_data()
        logger.debug(" - clean Xcom table")
        session.query(TaskInstance).filter(
            TaskInstance.task_id == ti.task_id,
            TaskInstance.dag_id == ti.dag_id,
            TaskInstance.execution_date == dr.execution_date).delete(synchronize_session='fetch')
        session.commit()
        logger.debug(" - clean TaskInstance table")
    session.query(DagRun).filter(
        DagRun.dag_id == dr.dag_id,
        DagRun.run_id == dr.run_id,
    ).delete(synchronize_session='fetch')
    session.commit()
    logger.debug(" - clean dag_run table")


def stop_tasks(dr):
    logger.debug(f"""Stop tasks for {dr.dag_id} - {dr.run_id}""")
    for ti in dr.get_task_instances():
        logger.debug(f"""process {ti.dag_id} - {ti.task_id} - {ti.execution_date} - {ti.pid}""")
        if ti.state == State.RUNNING:
            try:
                process = psutil.Process(ti.pid) if ti.pid else None
            except Exception:
                logger.debug(f" - cannot find process by PID {ti.pid}")
                process = None
            ti.set_state(State.FAILED)
            logger.debug(" - set state to FAILED")
            if process:
                logger.debug(f" - wait for process {ti.pid} to exit")
                try:
                    process.wait(timeout=TIMEOUT * 2)  # raises psutil.TimeoutExpired if timeout. Makes task fail -> DagRun fails
                except psutil.TimeoutExpired as e:
                    logger.debug(f" - Done waiting for process {ti.pid} to die")


def remove_tmp_data(dr):
    logger.debug(f"""Remove tmp data for {dr.dag_id} - {dr.run_id}""")
    tmp_folder_set = set()
    for ti in dr.get_task_instances():
        ti_xcom_data = ti.xcom_pull(task_ids=ti.task_id) # can be None
        if ti_xcom_data and "outdir" in ti_xcom_data:
            tmp_folder_set.add(ti_xcom_data["outdir"])
    for tmp_folder in tmp_folder_set:
        try:
            shutil.rmtree(tmp_folder)
            logger.debug(f"""Successfully removed {tmp_folder}""")
        except Exception as ex:
            logger.error(f"""Failed to delete temporary output directory {tmp_folder}\n {ex}""")


def clean_dag_run(**context):
    dag_id = context['dag_run'].conf['remove_dag_id']
    run_id = context['dag_run'].conf['remove_run_id']
    dr_list = DagRun.find(dag_id=dag_id, run_id=run_id)
    for dr in dr_list:
        stop_tasks(dr)
        remove_tmp_data(dr)
        clean_db(dr)
        


dag = DAG(dag_id="clean_dag_run",
          start_date=days_ago(1),
          on_failure_callback=dag_on_failure,
          on_success_callback=dag_on_success,
          schedule_interval=None)


run_this = PythonOperator(task_id='clean_dag_run',
                          python_callable=clean_dag_run,
                          provide_context=True,
                          dag=dag)





import os
import sys
import uuid
import json
import queue
import logging
import requests
import itertools
import threading
import socketserver

from time import sleep
from shutil import rmtree
from urllib.parse import urljoin
from collections import OrderedDict
from cwltest.utils import compare, CompareFail
from http.server import SimpleHTTPRequestHandler

from cwl_airflow.utilities.helpers import (
    load_yaml,
    get_dir,
    get_absolute_path,
    get_rootname,
    get_api_failure_reason,
    get_compressed
)
from cwl_airflow.utilities.cwl import (
    load_job,
    embed_all_runs,
    fast_cwl_load
)


def get_listener_thread(                                       # safe to kill when run as daemon
    results_queue,
    port,
    daemon
):
    httpd = socketserver.TCPServer(("127.0.0.1", port), CustomHandler)
    httpd.results_queue = results_queue                        # to have access to results_queue from CustomHandler through self.server.results_queue
    return threading.Thread(
        target=httpd.serve_forever,
        daemon=daemon
    )


def get_spinner_thread(daemon):                                # safe to kill when run as daemon
    return threading.Thread(target=spin, daemon=daemon)


def get_checker_thread(                                        # is not thread safe when writing to "suite_data"
    suite_data,
    results_queue,
    daemon
):
    return threading.Thread(target=check_result,
                            daemon=daemon,
                            kwargs={
                                "suite_data": suite_data,
                                "results_queue": results_queue
                            })


class CustomHandler(SimpleHTTPRequestHandler):

    def log_message(self, format, *args):  # to suppress logging on each POST 200 response sent
        pass

    def do_POST(self):
        self.send_response(200)
        self.end_headers()
        if "status" in self.path:
            return None
        payload = json.loads(
            self.rfile.read(
                int(self.headers["Content-Length"])
            ).decode("UTF-8")
        )["payload"]
        if "results" in payload or payload.get("state", None) == "failed":     # "results" can be {}, so we should check only if key is present, but not value
            self.server.results_queue.put({
                "run_id":  payload["run_id"],
                "dag_id":  payload["dag_id"],
                "results": payload.get("results", None)
            })


def spin():
    spinner = itertools.cycle(['-', '/', '|', '\\'])
    while True:
        sys.stdout.write(next(spinner))
        sleep(0.1)
        sys.stdout.flush()
        sys.stdout.write('\b')


def get_unfinished_runs(suite_data):
    """
    Selects only those items from "suite_data" that
    are not finished yet, so we can wait for their
    results
    """

    suite_data_unfinished = OrderedDict()
    for run_id, test_data in suite_data.items():
        if not test_data["finished"]:
            suite_data_unfinished[run_id] = test_data
    return suite_data_unfinished


def check_result(suite_data, results_queue):
    previous_unfinished_count = None
    while True:
        suite_data_unfinished = get_unfinished_runs(suite_data)
        if len(suite_data_unfinished) == 0:
            break
        try:
            item = results_queue.get(False)
        except queue.Empty:
            unfinished_count = len(suite_data_unfinished)
            if previous_unfinished_count != unfinished_count:
                logging.info(f"Waiting for {unfinished_count} unfinished runs:")
                for run_id, test_data in suite_data_unfinished.items():
                    logging.info(f"   test case - {test_data['index']}, dag_id - {test_data['dag_id']}, run_id - {run_id}")   
                previous_unfinished_count = unfinished_count
            sleep(10)                                       # sleep for 10 second before trying to fetch new results from the empty queue
            continue
        run_id = item["run_id"]
        test_data = suite_data_unfinished[run_id]  # if this fails, look for a bug
        logging.info(f"Check results from the test case {test_data['index']} that runs DAG {test_data['dag_id']} as {run_id}")
        try:
            compare(test_data["output"], item["results"])
        except (CompareFail, KeyError) as ex:               # catch KeyError in case output field is missing for tool that should fail
            if not test_data.get("should_fail", None):      # do not report error if tool should fail
                test_data["error"] = str(ex)
        finally:
            test_data["finished"] = True
            rmtree(test_data["job"]["outputs_folder"])


def load_test_suite(args):
    """
    Loads tests from the provided --suite file.
    Selects tests based on the indices from --range.
    
    Updates tools locations to be absolute, loads
    jobs and updates all inputs files locations to
    be absolute too. Adds "outputs_folder" to the job,
    as well as the "index" to indicate which test case
    was used.

    Adds run_id's as keys for easy access and proper
    test identification when receiving results.
    """

    suite_data = load_yaml(args.suite)
    suite_dir = os.path.dirname(args.suite)
    suite_data_filtered = OrderedDict()                                       # use OrderedDict just to keep it similar to suite_data
    for i in args.range:
        test_data = suite_data[i]
        run_id = str(uuid.uuid4())
        tool_location = get_absolute_path(test_data["tool"], suite_dir)
        logging.info(f"Read test case {i+1} to run {tool_location}")

        job_location = None
        job_data = {}

        if "job" in test_data:
            job_location = get_absolute_path(test_data["job"], suite_dir)
            job_data = load_job(
                workflow=tool_location,
                job=job_location
            )

        job_data["outputs_folder"] = get_dir(os.path.join(args.tmp, run_id))

        test_data.update({
            "job":  job_data,                                                 # already parsed, includes "outputs_folder"
            "tool": tool_location,
            "dag_id": get_rootname(test_data["tool"]),
            "index": i+1,                                                     # to know test case number, 1-based to correspond to --range
            "finished": False                                                 # to indicate whether the test was finished or not
        })
        logging.info(f"Successfully loaded test case {i+1} to run {tool_location} with {job_location} as {run_id}")
        suite_data_filtered[run_id] = test_data                               # use "run_id" as a key for fast access when checking results
    return suite_data_filtered


def create_dags(suite_data, args, dags_folder=None):
    """
    Iterates over "suite_data" and creates new DAGs. Tries to include
    all tools into the worfklow before sending it to the API server.
    If loaded tool is not Workflow, send it unchanged. It's safe to
    not process errors when we failed to add new DAG. Airflow Scheduler
    will parse all dags at the end of the next "dag_dir_list_interval"
    from airflow.cfg. If args.embed was True, send base64 encoded zlib
    compressed content of the workflow file instead of attaching it.
    """

    # TODO: Do we need to force scheduler to reload DAGs after all DAG added?

    for test_data in suite_data.values():
        workflow_path = os.path.join(
            args.tmp,
            os.path.basename(test_data["tool"])
        )
        embed_all_runs(                                                                               # will save results to "workflow_path"
            workflow_tool=fast_cwl_load(test_data["tool"]),
            location=workflow_path
        )
        with open(workflow_path, "rb") as input_stream:
            logging.info(f"Add DAG {test_data['dag_id']} from test case {test_data['index']}")

            if args.embed:                                                                            # send base64 encoded zlib compressed workflow content that will be embedded into DAG python file
                logging.info(f"Sending base64 encoded zlib compressed content from {workflow_path}")
                r = requests.post(
                    url=urljoin(args.api, "/api/experimental/dags"),
                    params={
                        "dag_id": test_data["dag_id"]
                    },
                    json={"workflow_content": get_compressed(input_stream)}
                )
            else:                                                                                     # attach workflow as a file
                logging.info(f"Attaching workflow file {workflow_path}")
                r = requests.post(
                    url=urljoin(args.api, "/api/experimental/dags"),
                    params={
                        "dag_id": test_data["dag_id"]
                    },
                    files={"workflow": input_stream}
                )

            # Check if we failed to add new DAG. One reason to fail - DAG hase been
            # already added. It's safe to ignore this error. In case more serious
            # reasons, they will be caught on the "trigger_dags" step

            if not r.ok:
                reason = get_api_failure_reason(r)
                logging.error(f"Failed to add DAG {test_data['dag_id']} from test case {test_data['index']} due to \n {reason}")


def trigger_dags(suite_data, args):
    """
    Triggers all DAGs from "suite_data". If failed to trigger DAG, updates
    "suite_data" with "error" and sets "finished" to True
    """

    for run_id, test_data in suite_data.items():
        logging.info(f"Trigger DAG {test_data['dag_id']} from test case {test_data['index']} as {run_id}")
        r = requests.post(
            url=urljoin(args.api, "/api/experimental/dag_runs"),
            params={
                "run_id": run_id,
                "dag_id": test_data["dag_id"],
                "conf": json.dumps(
                    {
                        "job": test_data["job"]
                    }
                )
            }
        )
        if not r.ok:
            reason = get_api_failure_reason(r)
            logging.error(f"Failed to trigger DAG {test_data['dag_id']} from test case {test_data['index']} as {run_id} due to {reason}")
            test_data["error"] = reason
            test_data["finished"] = True


def print_report(suite_data):
    exit_code = 0
    for run_id, test_data in suite_data.items():  # no need to check if "finished", because all items shoud be finished at this step
        if "error" in test_data:
            exit_code = 1
            logging.error(f"Test case {test_data['index']} that runs DAG {test_data['dag_id']} as {run_id} failed with error \n{test_data['error']}")
        else:
            logging.info(f"Test case {test_data['index']} that runs DAG {test_data['dag_id']} as {run_id} finished successfully")
    return exit_code


def run_test_conformance(args):
    """
    Runs conformance tests
    """

    # TODO: do not forget to remove args.tmp

    # Load test suite data, setup a queue to keep results
    suite_data = load_test_suite(args)
    results_queue = queue.Queue(maxsize=len(suite_data))

    # Create new dags
    create_dags(suite_data, args)                           # only reads from "suite_data"

    # Start thread to listen for status updates before
    # we trigger DAGs. "results_queue" is thread safe
    listener = get_listener_thread(
        results_queue=results_queue,
        port=args.port,
        daemon=True
    )
    listener.start()

    # Trigger all dags updating "suite_data" items with "error" and "finished"=True
    # for all DAG runs that we failed to trigger. Writing to "suite_data" is not
    # thread safe!
    trigger_dags(suite_data, args)

    # Start checker thread to evaluate received results.
    # Writes to "suite_data" which is not thread safe,
    # that's why we start thread after we triggered all DAGs.
    checker = get_checker_thread(
        suite_data=suite_data,
        results_queue=results_queue,
        daemon=False
    )
    checker.start()

    # Display spinner if --spin
    if args.spin:
        spinner = get_spinner_thread(daemon=True)
        spinner.start()

    # Wait until no unfinished items left in "suite_data"
    checker.join()

    exit_code = print_report(suite_data)

    sys.exit(exit_code)


import os
import re
import sys
import dill as pickle  # standard pickle doesn't handle lambdas
import argparse
import json
import zlib
import shutil
import docker
import logging
import binascii

from copy import deepcopy
from jsonmerge import merge
from urllib.parse import urlsplit
from tempfile import NamedTemporaryFile
from typing import MutableMapping, MutableSequence
from ruamel.yaml.comments import CommentedMap
from airflow.configuration import (
    AIRFLOW_HOME,
    conf
)
from airflow.exceptions import AirflowConfigException
from cwltool.argparser import get_default_args
from cwltool.main import (
    setup_loadingContext,
    init_job_order
)
from cwltool.context import (
    LoadingContext,
    RuntimeContext
)
from cwltool.process import relocateOutputs
from cwltool.load_tool import (
    load_tool,
    jobloaderctx
)
from cwltool.executors import SingleJobExecutor
from cwltool.utils import visit_class
from cwltool.mutation import MutationManager
from schema_salad.ref_resolver import Loader
from schema_salad.exceptions import SchemaSaladException
from schema_salad.ref_resolver import file_uri

from cwl_airflow.utilities.helpers import (
    get_md5_sum,
    get_dir,
    load_yaml,
    dump_json,
    get_absolute_path,
    get_rootname,
    remove_field_from_dict,
    get_uncompressed,
    get_compressed,
    get_files
)


CWL_TMP_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_tmp_folder")
CWL_OUTPUTS_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_outputs_folder")
CWL_INPUTS_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_inputs_folder")
CWL_PICKLE_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_pickle_folder")
CWL_USE_CONTAINER = True
CWL_NO_MATCH_USER = False
CWL_SKIP_SCHEMAS = True
CWL_STRICT = False
CWL_QUIET = False
CWL_RM_TMPDIR = True       # better not to change without need
CWL_MOVE_OUTPUTS = "move"  # better not to change without need
CWL_ENABLE_DEV = True      # better not to change without need


DAG_TEMPLATE="""#!/usr/bin/env python3
from cwl_airflow.extensions.cwldag import CWLDAG
dag = CWLDAG(
    workflow="{0}",
    dag_id="{1}"
)
"""


def overwrite_deprecated_dag(
    dag_location,
    deprecated_dags_folder=None
):
    """
    Loads DAG content from "dag_location" file. Searches for "dag.create()" command.
    If not found, we don't need to upgrade this DAG (it's either not from CWL-Airflow,
    or already in a new format). If "deprecated_dags_folder" is not None, copies original
    DAG file there before DAG upgrading. After copying deprecated DAG to the
    "deprecated_dags_folder" updates ".airflowignore" with DAG file basename to exclude
    it from Airflow parsing. Upgraded DAG will always include base64 encoded zlib
    compressed workflow content. In case "workflow_location" is relative path, it will
    be resolved based on the dirname of "dag_location" (useful for tests only, because
    all our old DAGs always have absolute path to the CWL file). Function doesn't backup
    or update the original CWL file.
    TODO: in case more coplicated DAG files that include "default_args", etc, this function
    should be updated to the more complex one.
    """

    with open(dag_location, "r+") as io_stream:                # open for both reading and writing

        dag_content = io_stream.read()

        if not re.search("dag\\.create\\(\\)", dag_content):   # do nothing if it wasn't old-style DAG
            return

        workflow_location = get_absolute_path(                 # resolve relative to dirname of "dag_location" (good for tests)
            re.search(
                "(cwl_workflow\\s*=\\s*[\"|'])(.+?)([\"|'])",
                dag_content
            ).group(2),
            os.path.dirname(dag_location)
        )

        dag_id = re.search(
            "(dag_id\\s*=\\s*[\"|'])(.+?)([\"|'])",
            dag_content
        ).group(2)

        compressed_workflow_content = get_compressed(
            fast_cwl_load(workflow_location)                   # no "run" embedding or convertion to Workflow. If DAG worked, cwl should be ok too
        )

        if deprecated_dags_folder is not None:                 # copy old DAG to the folder with deprecated DAGs, add ".airflowignore"
            get_dir(deprecated_dags_folder)                    # try to create "deprecated_dags_folder" if it doesn't exist
            shutil.copy(dag_location, deprecated_dags_folder)  # copy DAG file
            ignore = os.path.join(
                deprecated_dags_folder,
                ".airflowignore"
            )
            with open(ignore, "a") as output_stream:           # add deprecated DAG to ".airflowignore"
                output_stream.write(
                    os.path.basename(dag_location) + "\n"
                )

        io_stream.seek(0)                                      # rewind "dag_location" file to the beginning
        io_stream.write(
            DAG_TEMPLATE.format(
                compressed_workflow_content, dag_id
            )
        )
        io_stream.truncate()                                   # remove old data at the end of a file if anything became shorter than original


def conf_get(
    section,
    key,
    default
):
    """
    Return value from AirflowConfigParser object.
    If section or key is absent, return default.
    Suppresses annoying warning messages
    """

    try:
        logging.disable(logging.WARNING)
        return conf.get(section, key)
    except AirflowConfigException:
        return default
    finally:
        logging.disable(logging.NOTSET)  # is guaranteed to be executed before any return


def collect_reports(
    context,
    task_ids=None
):
    """
    Collects reports from "context" for specified "task_ids".
    If "task_ids" was not set, use all tasks from DAG. Loads
    and merges data from reports.
    """

    task_ids = context["dag"].task_ids if task_ids is None else task_ids

    job_data = {}
    for report_location in context["ti"].xcom_pull(task_ids=task_ids):
        if report_location is not None:
            job_data = merge(
                job_data,
                load_yaml(report_location)
            )
    
    return job_data


def get_default_cwl_args(preset_cwl_args=None):
    """
    Returns default arguments required by cwltool's functions with a few
    parameters added and overwritten (required by CWL-Airflow). Defaults
    can be preset through "preset_cwl_args" if provided. All new fields
    from "preset_cwl_args" will be added to the returned results.
    """

    preset_cwl_args = {} if preset_cwl_args is None else deepcopy(preset_cwl_args)

    # default arguments required by cwltool
    required_cwl_args = get_default_args()

    # update default arguments required by cwltool with those that were preset by user
    required_cwl_args.update(preset_cwl_args)

    # update default arguments required by cwltool with those that might
    # be updated based on the higher priority of airflow configuration
    # file. If airflow configuration file doesn't include correspondent
    # parameters, use those that were preset by user, or defaults
    required_cwl_args.update(
        {
            "tmp_folder": get_dir(
                conf_get(
                    "cwl", "tmp_folder",
                    preset_cwl_args.get("tmp_folder", CWL_TMP_FOLDER)
                )
            ),
            "outputs_folder": get_dir(                                             # for CWL-Airflow to store outputs if "outputs_folder" is not overwritten in job
                conf_get(
                    "cwl", "outputs_folder",
                    preset_cwl_args.get("outputs_folder", CWL_OUTPUTS_FOLDER)
                )
            ),
            "inputs_folder": get_dir(                                             # for CWL-Airflow to resolve relative locations for input files if job was loaded from parsed object
                conf_get(
                    "cwl", "inputs_folder",
                    preset_cwl_args.get("inputs_folder", CWL_INPUTS_FOLDER)
                )
            ),
            "pickle_folder": get_dir(                                              # for CWL-Airflow to store pickled workflows
                conf_get(
                    "cwl", "pickle_folder",
                    preset_cwl_args.get("pickle_folder", CWL_PICKLE_FOLDER)
                )
            ),
            "use_container": conf_get(
                "cwl", "use_container",
                preset_cwl_args.get("use_container", CWL_USE_CONTAINER)            # execute jobs in docker containers
            ),
            "no_match_user": conf_get(
                "cwl", "no_match_user",
                preset_cwl_args.get("no_match_user", CWL_NO_MATCH_USER)            # disables passing the current uid to "docker run --user"
            ),
            "skip_schemas": conf_get(
                "cwl", "skip_schemas", 
                preset_cwl_args.get("skip_schemas", CWL_SKIP_SCHEMAS)              # it looks like this doesn't influence anything in the latest cwltool
            ),
            "strict": conf_get(
                "cwl", "strict", 
                preset_cwl_args.get("strict", CWL_STRICT)
            ),
            "quiet": conf_get(
                "cwl", "quiet", 
                preset_cwl_args.get("quiet", CWL_QUIET)
            ),
            "rm_tmpdir": preset_cwl_args.get("rm_tmpdir", CWL_RM_TMPDIR),          # even if we can set it in "preset_cwl_args" it's better not to change
            "move_outputs": preset_cwl_args.get("move_outputs", CWL_MOVE_OUTPUTS), # even if we can set it in "preset_cwl_args" it's better not to change
            "enable_dev": preset_cwl_args.get("enable_dev", CWL_ENABLE_DEV)        # fails to run without it when creating workflow from tool. TODO: Ask Peter?
        }
    )

    return required_cwl_args


def relocate_outputs(
    workflow,
    job_data,
    cwl_args=None,
    remove_tmp_folder=None
):
    """
    Relocates filtered outputs to "outputs_folder" and, by default,
    removes tmp_folder, unless "remove_tmp_folder" is set to something
    else. Saves report with relocated outputs as "workflow_report.json"
    to "outputs_folder".
    Maps outputs from "workflow" back to normal (from step_id_step_out
    to workflow output) and filters "job_data" based on them (combining
    items from "job_data" into a list based on "outputSource" if it
    was a list). "cwl_args" can be used to update default parameters
    used for loading and runtime contexts.
    """

    cwl_args = {} if cwl_args is None else cwl_args
    remove_tmp_folder = True if remove_tmp_folder is None else remove_tmp_folder

    default_cwl_args = get_default_cwl_args(cwl_args)

    workflow_tool = fast_cwl_load(
        workflow=workflow,
        cwl_args=default_cwl_args
    )

    # Filter "job_data" to include only items required by workflow outputs.
    # Remap keys to the proper workflow outputs IDs (without step id).
    # If "outputSource" was a list even of len=1, find all correspondent items
    # from the "job_data" and assign them as list of the same size. 
    job_data_copy = deepcopy(job_data)
    filtered_job_data = {}
    for output_id, output_data in get_items(workflow_tool["outputs"]):
        collected_job_items = []
        for source_id, _ in get_items(output_data["outputSource"]):
            collected_job_items.append(
                job_data_copy[source_id.replace("/", "_")]
            )
        if isinstance(output_data["outputSource"], list):
            filtered_job_data[output_id] = collected_job_items
        else:
            filtered_job_data[output_id] = collected_job_items[0]

    # Outputs will be always copied, because source_directories=[]
    runtime_context = RuntimeContext(default_cwl_args)
    relocated_job_data = relocateOutputs(
        outputObj=filtered_job_data,
        destination_path=job_data_copy["outputs_folder"],
        source_directories=[],                              # use it as a placeholder (shouldn't influence anything)
        action=runtime_context.move_outputs,
        fs_access=runtime_context.make_fs_access(""),
        compute_checksum=runtime_context.compute_checksum,
        path_mapper=runtime_context.path_mapper
    )

    # Dump report with relocated outputs
    workflow_report = os.path.join(
        job_data_copy["outputs_folder"],
        "workflow_report.json"
    )

    dump_json(relocated_job_data, workflow_report)

    # Clean "tmp_folder"
    if remove_tmp_folder:
        shutil.rmtree(job_data_copy["tmp_folder"], ignore_errors=False)

    return relocated_job_data, workflow_report


def get_containers(job_data, task_id):
    """
    Searches for cidfiles in the "step_tmp_folder", loads
    container IDs from the found files, adds them to dict
    in a form of {cid: location}. If nothing found,
    returns {}.
    """

    containers = {}

    step_tmp_folder, _, _, _ = get_temp_folders(
        task_id=task_id,
        job_data=job_data
    )

    for location in get_files(step_tmp_folder, ".*\\.cid$").values():
        try:
            with open(location, "r") as input_stream:
                containers[input_stream.read()] = location
        except OSError as err:
            logging.error(f"Failed to read container ID \
                from {location} due to \n{err}")

    return containers


def kill_containers(containers):
    """
    Iterates over "containers" dictionary received from "get_containers"
    and tries to kill all running containers based on cid. If killed
    container was not in "running" state, was successfully killed or not
    found at all, removes correspondent cidfile.
    """

    docker_client = docker.from_env()
    for cid, location in containers.items():
        try:
            container = docker_client.containers.get(cid)
            if container.status == "running":
                container.kill()
            os.remove(location)
        except docker.errors.NotFound as err:
            os.remove(location)
        except docker.errors.APIError as err:
            logging.error(f"Failed to kill container. \n {err}")


def execute_workflow_step(
    workflow,
    task_id,
    job_data,
    cwl_args=None,
    executor=None
):
    """
    Constructs and executes single step workflow based on the "workflow"
    and "task_id". "cwl_args" can be used to update default parameters
    used for loading and runtime contexts. Exports json file with the
    execution results.
    """

    cwl_args = {} if cwl_args is None else cwl_args
    executor = SingleJobExecutor() if executor is None else executor

    step_tmp_folder, step_cache_folder, step_outputs_folder, step_report = get_temp_folders(
        task_id=task_id,
        job_data=job_data
    )

    default_cwl_args = get_default_cwl_args(cwl_args)

    default_cwl_args.update({                          # add execution specific parameters
        "tmp_outdir_prefix": step_cache_folder + "/",
        "tmpdir_prefix": step_cache_folder + "/",
        "cidfile_dir": step_tmp_folder,
        "cidfile_prefix": task_id,
        "basedir": os.getcwd(),                        # job should already have abs path for inputs, so this is useless
        "outdir": step_outputs_folder
    })
    
    workflow_step_path = os.path.join(
        step_tmp_folder, task_id + "_step_workflow.cwl"
    )

    fast_cwl_step_load(                                # will save new worlflow to "workflow_step_path"
        workflow=workflow,
        target_id=task_id,
        cwl_args=default_cwl_args,
        location=workflow_step_path
    )

    _stderr = sys.stderr                               # to trick the logger	
    sys.stderr = sys.__stderr__
    step_outputs, step_status = executor(
        slow_cwl_load(
            workflow=workflow_step_path,
            cwl_args=default_cwl_args),
        job_data,
        RuntimeContext(default_cwl_args)
    )
    sys.stderr = _stderr

    if step_status != "success":
        raise ValueError

    # To remove "http://commonwl.org/cwltool#generation": 0 (copied from cwltool)
    visit_class(step_outputs, ("File",), MutationManager().unset_generation)

    dump_json(step_outputs, step_report)

    return step_outputs, step_report


def get_temp_folders(task_id, job_data):
    """
    Creates a set of folders required for workflow execution.
    Uses "tmp_folder" from "job_data" as a parent folder.
    """

    step_tmp_folder = get_dir(
        os.path.join(
            job_data["tmp_folder"],
            task_id
        )
    )

    step_cache_folder = get_dir(
        os.path.join(
            step_tmp_folder,
            task_id + "_step_cache"
        )
    )


    step_outputs_folder = get_dir(
        os.path.join(
            step_tmp_folder,
            task_id + "_step_outputs"
        )
    )

    step_report = os.path.join(
        step_tmp_folder,
        task_id + "_step_report.json"
    )

    return step_tmp_folder, step_cache_folder, step_outputs_folder, step_report


def load_job(
    workflow,
    job,
    cwl_args=None,
    cwd=None
):
    """
    Tries to load json object from "job". If failed, assumes that
    "job" has been already parsed into Object. Inits loaded "job_data"
    based on the "workflow" (mostly for setting defaults from the workflow
    inputs; never fails). "cwl_args" can be used to update parameters for
    loading and runtime contexts.

    If "job" was file, resolves relative paths based on the job file location.
    If "job" was already parsed into Object, resolves relative paths based on
    "cwd". If "cwd" was None uses "inputs_folder" value from "cwl_args" or
    its default value returned from "get_default_cwl_args" function.

    Checking links after relative paths are resolved is disabled (checklinks
    is set to False in both places). This will prevent rasing an exception by
    schema salad in those cases when an input file will be created from the
    provided content during workflow execution.
    
    Always returns CommentedMap
    """
    
    cwl_args = {} if cwl_args is None else cwl_args
    
    default_cwl_args = get_default_cwl_args(cwl_args)
    cwd = default_cwl_args["inputs_folder"] if cwd is None else cwd

    loading_context = setup_loadingContext(
        LoadingContext(default_cwl_args),
        RuntimeContext(default_cwl_args),
        argparse.Namespace(**default_cwl_args)
    )

    job_copy = deepcopy(job)

    try:
        job_data, _ = loading_context.loader.resolve_ref(job_copy, checklinks=False)
    except (FileNotFoundError, SchemaSaladException) as err:
        job_data = load_yaml(json.dumps(job_copy))
        job_data["id"] = file_uri(cwd) + "/"
        job_data, metadata = loading_context.loader.resolve_all(
            job_data,
            job_data["id"],
            checklinks=False
        )

    initialized_job_data = init_job_order(
        job_order_object=job_data,
        args=argparse.Namespace(**default_cwl_args),
        process=slow_cwl_load(
            workflow=workflow, 
            cwl_args=default_cwl_args
        ),
        loader=loading_context.loader,
        stdout=os.devnull
    )

    return initialized_job_data


def fast_cwl_step_load(workflow, target_id, cwl_args=None, location=None):
    """
    Returns workflow (CommentedMap) that includes only single step
    selected by "target_id" from the parsed "workflow". Other steps
    are removed. Workflow inputs and outputs are updated based on
    source fields of "in" and "out" from the selected workflow step.
    If selected step includes "scatter" field all output types will
    be transformed to the nested/flat array of items of the same type. 
    IDs of updated workflow inputs and outputs as well as IDs of
    correspondent "source" fields also include step id separated by
    underscore. All other fields remain unchanged.

    "cwl_args" can be used to update default location of "pickle_folder"
    used by "fast_cwl_load" as well as other parameters used by
    "slow_cwl_load" for loading and runtime contexts.

    If "location" is not None, export modified workflow.
    """

    cwl_args = {} if cwl_args is None else cwl_args

    default_cwl_args = get_default_cwl_args(cwl_args)

    workflow_inputs = []
    workflow_outputs = []
    workflow_steps = []

    workflow_tool = fast_cwl_load(
        workflow=workflow,
        cwl_args=default_cwl_args
    )

    selected_step = list(get_items(workflow_tool["steps"], target_id))[0][1]

    workflow_steps.append(selected_step)

    for _, step_in in get_items(selected_step.get("in", [])):           # step might not have "in"
        
        updated_sources = []  # to keep track of updated sources

        for step_in_source, _ in get_items(step_in.get("source", [])):  # "in" might not have "source"

            try:

                # try to find workflow input that corresponds to "source"

                workflow_input = list(get_items(
                    workflow_tool["inputs"],
                    step_in_source
                ))[0][1]

                updated_workflow_input = {
                    "id": step_in_source,
                    "type": workflow_input["type"]
                }

                # need to copy:
                #  original inputBinding because it can include loadContents section
                #  loadContents and loadListing sections if present outside of inputBinding
                #  both "default" and "secondaryFiles" if present
                # TODO: Do I need to copy format?
                for key in ["default", "secondaryFiles", "inputBinding", "loadContents", "loadListing"]:
                    if key in workflow_input:
                        updated_workflow_input[key] = workflow_input[key]

                # Check if we have already added input based on the same "source"
                # from another item from "in". Skip adding the same input twice.

                if len(list(get_items(workflow_inputs, step_in_source))) == 0:
                    workflow_inputs.append(updated_workflow_input)

                updated_sources.append(step_in_source)

            except (IndexError, KeyError):
                
                # Need to find upstream step that corresponds to "source"

                upstream_step = list(get_items(
                    workflow_tool["steps"],
                    get_short_id(step_in_source, only_step_name=True)
                ))[0][1]

                # Need to load tool from "run" of the found upstream step
                # and look for the output that corresponds to "source".
                # We look for correspondence only based on "id"

                upstream_step_tool = fast_cwl_load(
                    workflow=upstream_step["run"],
                    cwl_args=default_cwl_args
                )

                upstream_step_output = list(get_items(
                    {get_short_id(k, only_id=True): v for k, v in get_items(upstream_step_tool["outputs"])},  # trick
                    get_short_id(step_in_source, only_id=True)
                ))[0][1]

                step_in_source_with_step_id = step_in_source.replace("/", "_")  # to include both step name and id

                # Check if it should be assumed optional (default field is present)
                # NOTE: consider also checking if upstream step had scatter, so the
                # output type should become array based on the scatter parameters
                if "default" in step_in:
                    upstream_step_output_type = ["null", upstream_step_output["type"]]
                else:
                    upstream_step_output_type = upstream_step_output["type"]

                updated_workflow_input = {
                    "id": step_in_source_with_step_id,  
                    "type": upstream_step_output_type
                }

                # No need to copy "secondaryFiles" for outputs from other steps
                # because they should be already included into the generated json
                # report file
                # # TODO: Do I need to copy format to "workflow_inputs"?

                # Check if we have already added input based on the same "source"
                # from another item from "in". Skip adding the same input twice.

                if len(list(get_items(workflow_inputs, step_in_source_with_step_id))) == 0:
                    workflow_inputs.append(updated_workflow_input)
                
                updated_sources.append(step_in_source_with_step_id)

        # replace "source" in step's "in" if anything was updated
        if len(updated_sources) > 0:
            if isinstance(step_in["source"], list):
                step_in["source"] = updated_sources
            else:
                step_in["source"] = updated_sources[0]   

    # Need to load tool from the "run" field of the selected step
    # and look for the outputs that correspond to the items from "out".
    # We look for correspondence only based on "id"

    selected_step_tool = fast_cwl_load(
        workflow=selected_step["run"],
        cwl_args=default_cwl_args
    )

    for step_out, _ in get_items(selected_step["out"]):
        selected_step_output = list(get_items(
            {get_short_id(k, only_id=True): v for k, v in get_items(selected_step_tool["outputs"])},  # trick
            get_short_id(step_out, only_id=True)
        ))[0][1]
        step_out_with_step_id = step_out.replace("/", "_")  # to include both step name and id

        # update output type in case of scatter
        if "scatter" in selected_step:
            selected_step_output = deepcopy(selected_step_output)                  # need to deepcopy, otherwise we change embedded tool's output
            if isinstance(selected_step["scatter"], MutableSequence) \
                and selected_step.get("scatterMethod") == "nested_crossproduct":
                nesting = len(selected_step["scatter"])
            else:
                nesting = 1
            for _ in range(0, nesting):
                selected_step_output["type"] = {
                    "type": "array",
                    "items": selected_step_output["type"]
                }

        workflow_outputs.append({
            "id": step_out_with_step_id,
            "type": selected_step_output["type"],
            "outputSource": step_out
        })

    workflow_tool.update(
        {
            "inputs": workflow_inputs,
            "outputs": workflow_outputs,
            "steps": workflow_steps
        }
    )

    if location is not None:
        dump_json(workflow_tool, location)

    return workflow_tool


def get_items(data, target_id=None):
    """
    If data is dictionary returns [(key, value)].
    
    If data is string return [key, data].
    
    If data is list of items returns [(key, value)] with key set
    from item["id"] and value equal to the item itself.

    If items are strings, set keys from this strings and return [(key, item)].
    
    For all other cases return either list of tuples of unchanged items or
    tuple with unchanged input data.
    
    Keys are always shortened to include only part after symbol #

    If target_id is set, filter outputs by key == target_id if
    found. Otherwise return []
    """

    if isinstance(data, MutableMapping):
        for key, value in data.items():
            if target_id is not None:
                if key == target_id or get_short_id(key) == target_id:
                    yield get_short_id(key), value 
                else:
                    continue
            else:
                yield get_short_id(key), value 
    elif isinstance(data, MutableSequence):
        for item in data:
            if isinstance(item, str):
                if target_id is not None:
                    if item == target_id or get_short_id(item) == target_id:
                        yield get_short_id(item), item
                    else:
                        continue
                else:
                    yield get_short_id(item), item
            elif "id" in item:  # we checked that item wasn't string, so "id" is not substring of "item"
                if target_id is not None:
                    if item["id"] == target_id or get_short_id(item["id"]) == target_id:
                        yield get_short_id(item["id"]), item
                    else:
                        continue
                else:
                    yield get_short_id(item["id"]), item
            else:
                if target_id is not None:
                    if item == target_id:
                        yield item, item
                    else:
                        pass  # do not yield anything
                else:
                    yield item, item
    elif isinstance(data, str):
        if target_id is not None:
            if data == target_id or get_short_id(data) == target_id:
                yield get_short_id(data), data
            else:
                pass  # do not yield anything
        else:
            yield get_short_id(data), data
    else:
        if target_id is not None:
            if data == target_id:
                yield data, data
            else:
                pass  # do not yield anything
        else:
            yield data, data


def get_short_id(long_id, only_step_name=None, only_id=None):
    """
    Shortens long id to include only part after symbol #.
    If # is not present, use long_id. If only_step_name is True,
    return a short step name. If only_id is True, return a short
    id without step name. If part after symbol # includes three
    or more sections separated by "/", discard the middle ones.
    If part after symbol # includes only one section separated by
    "/", reset "only_step_name" and "only_id" to False as we don't
    know whether it was step name of "id"
    """
    fragment = urlsplit(long_id).fragment
    part = fragment if fragment != "" else long_id

    if len(part.split("/")) > 2:                                  # if "id" has weird stuff between step name and id
        part = "/".join([part.split("/")[0], part.split("/")[-1]])

    if len(part.split("/")) == 1:                                  # if fragment is only one word (based on "/")
        only_step_name, only_id = False, False

    part = part.split("/")[0] if only_step_name else part
    part = "/".join(part.split("/")[1:]) if only_id else part
    return part


def fast_cwl_load(workflow, cwl_args=None):
    """
    Tries to unpickle workflow from "pickle_folder" based on
    md5 sum of the "workflow" file. "cwl_args" can be used to update
    default location of "pickle_folder" as well as other parameters
    used by "slow_cwl_load" for loading and runtime contexts.
    
    If pickled file not found or failed to unpickle, load tool from
    the "workflow" using "slow_cwl_load" with "only_tool" set to True
    to return only tool. Returned tool will be pickled into "pickle_folder"
    with a basename generated from md5 sum of the "workflow" file.

    If "workflow" was already parsed into CommentedMap, return it unchanged.
    Nothing will be pickled
    """

    cwl_args = {} if cwl_args is None else cwl_args

    if isinstance(workflow, CommentedMap):
        return workflow

    default_cwl_args = get_default_cwl_args(cwl_args)

    pickled_workflow = os.path.join(
        default_cwl_args["pickle_folder"],
        get_md5_sum(workflow) + ".p"
    )

    try:

        with open(pickled_workflow, "rb") as input_stream:
            workflow_tool = pickle.load(input_stream)

    except (FileNotFoundError, pickle.UnpicklingError) as err:

        workflow_tool = slow_cwl_load(
            workflow=workflow,
            cwl_args=default_cwl_args,
            only_tool=True
        )

        with open(pickled_workflow , "wb") as output_stream:
            pickle.dump(workflow_tool, output_stream)

    return workflow_tool


def slow_cwl_load(workflow, cwl_args=None, only_tool=None):
    """
    Follows standard routine for loading CWL file the same way
    as cwltool does it. "workflow" should be an absolute path
    the cwl file to load. "cwl_args" can be used to update
    default arguments mostly used for loading and runtime contexts.
    If "only_tool" is True, return only tool (used for pickling,
    because the whole Workflow object later fails to be unpickled).
    If "workflow" was already parsed into CommentedMap, return it
    unchanged (in a form similar to what we can get if parsed
    "workflow" with "only_tool" set to True).
    If "workflow" was a zlib compressed content of a file, it needs
    to be uncompressed, then written to the temp file and loaded the
    same way as described above. After loading temp file will be
    removed automatically. First always try to uncompress, because
    it's faster.
    """
    
    cwl_args = {} if cwl_args is None else cwl_args
    only_tool = False if only_tool is None else only_tool

    if isinstance(workflow, CommentedMap):
        return workflow

    default_cwl_args = get_default_cwl_args(cwl_args)

    def __load(location):
        return load_tool(
            location,
            setup_loadingContext(
                LoadingContext(default_cwl_args),
                RuntimeContext(default_cwl_args),
                argparse.Namespace(**default_cwl_args)
            )
        )

    try:
        with NamedTemporaryFile(mode="w") as temp_stream:  # guarantees that temp file will be removed
            json.dump(
                get_uncompressed(workflow, parse_as_yaml=True),
                temp_stream
            )
            temp_stream.flush()                            # otherwise it might be only partially written
            workflow_data = __load(temp_stream.name)
    except (zlib.error, binascii.Error):                   # file was real
        workflow_data = __load(workflow)

    return workflow_data.tool if only_tool else workflow_data


def embed_all_runs(
    workflow_tool,
    cwl_args=None,
    location=None
):
    """
    Tries to find and load all "run" fields from the "workflow_tool"
    if it is Workflow. If not, doesn't replace anything. "cwl_args"
    can be used to update default arguments used by loading and runtime
    contexts. If "location" is provided, save resulted workflow to json
    file. Returns workflow tool with all "run" fields replaced.
    """

    def __embed(workflow_tool, cwl_args=None):
        if isinstance(workflow_tool, MutableSequence):
            for item in workflow_tool:
                __embed(item, cwl_args)
        elif isinstance(workflow_tool, MutableMapping):
            if "run" in workflow_tool and isinstance(workflow_tool["run"], str):
                workflow_tool["run"] = slow_cwl_load(
                    workflow=workflow_tool["run"],
                    cwl_args=cwl_args,
                    only_tool=True)
            for item in workflow_tool.values():
                __embed(item, cwl_args)

    if workflow_tool["class"] == "Workflow":
        workflow_tool_copy = deepcopy(workflow_tool)
        __embed(workflow_tool_copy, cwl_args)
    else:
        workflow_tool_copy = workflow_tool

    if location is not None:
        dump_json(workflow_tool_copy, location)

    return workflow_tool_copy


def convert_to_workflow(command_line_tool, location=None):
    """
    Converts "command_line_tool" to Workflow trying to keep all
    important elements. If "command_line_tool" is already Workflow,
    doesn't apply any changes. If "location" is not None, dumps
    results to json file.
    """

    if command_line_tool["class"] == "Workflow":
        workflow_tool = command_line_tool
    else:
        workflow_tool = {
            "class": "Workflow",
            "cwlVersion": command_line_tool["cwlVersion"],
            "inputs": [],
            "outputs": []
        }

        for key in ["requirements"]:
            if key in command_line_tool:
                workflow_tool[key] = command_line_tool[key]

        for input_id, input_data in get_items(command_line_tool["inputs"]):
            workflow_input = {
                "id": input_id,
                "type": remove_field_from_dict(input_data["type"], "inputBinding")       # "type" in WorkflowInputParameter cannot have "inputBinding"
            }
            for key in ["secondaryFiles", "default"]:  # TODO: Do I need to copy format?
                if key in input_data:
                    workflow_input[key] = input_data[key]
            workflow_tool["inputs"].append(workflow_input)

        for output_id, output_data in get_items(command_line_tool["outputs"]):
            workflow_output = {
                "id": output_id,
                "type": output_data["type"],
                "outputSource": get_rootname(command_line_tool["id"]) + "/" + output_id
            }
            # TODO: not sure if I need format here
            # for key in ["format"]:
            #     if key in output_data:
            #         workflow_output[key] = output_data[key]
            workflow_tool["outputs"].append(workflow_output)

        workflow_tool["steps"] = [
            {
                "id": get_rootname(command_line_tool["id"]),
                "run": command_line_tool,
                "in": [
                    {
                        "id": input_id, "source": input_id
                    } for input_id, _ in get_items(workflow_tool["inputs"])
                ],
                "out": [
                    output_id for output_id, _ in get_items(workflow_tool["outputs"])
                ]
            }
        ]

    if location is not None:
        dump_json(workflow_tool, location)

    return workflow_tool



import logging
import connexion
import random
import string
import json
import tempfile
import psutil
import shutil

import cwltool.load_tool as load  # TODO: use my functions instead
from schema_salad.ref_resolver import Loader

from time import sleep
from werkzeug.utils import secure_filename
from six import iterlists
from os import path, environ
from subprocess import check_call, CalledProcessError, DEVNULL

from airflow.settings import DAGS_FOLDER, AIRFLOW_HOME
from airflow.models import DagBag, TaskInstance, DagRun, DagModel
from airflow.utils.state import State
from airflow.utils.timezone import parse as parsedate
from airflow.utils.db import provide_session

from cwl_airflow.utilities.helpers import (
    get_version,
    get_dir,
    get_compressed,
    get_md5_sum
)
from cwl_airflow.utilities.cwl import (
    conf_get,
    fast_cwl_load,
    slow_cwl_load,
    convert_to_workflow,
    DAG_TEMPLATE
)


class CWLApiBackend():

    # curl -X GET "127.0.0.1:8081/wes/v1/dags" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dags?dag_ids=example_bash_operator" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dags?dag_ids=example_bash_operator,tutorial" -H "accept: application/json"

    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&state=running" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&state=success" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?state=running" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs" -H "accept: application/json"

    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?run_id=scheduled__2019-07-20T00%3A00%3A00%2B00%3A00" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?run_id=scheduled__2019-07-20T00%3A00%3A00%2B00%3A00&dag_id=tutorial" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?execution_date=2019-07-20T00%3A00%3A00%2B00%3A00" -H "accept: application/json"

    # curl -X POST "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&run_id=1234567&conf=%22%7B%7D%22" -H "accept: application/json"
    # curl -X POST "localhost:8081/wes/v1/dags?dag_id=bowtie2-indices" -H "accept: application/json" -H "Content-Type: multipart/form-data" -F "workflow=@bowtie2-indices.cwl"
    # curl -X POST "localhost:8081/wes/v1/dags/bowtie2-indices/dag_runs?run_id=bowtie2_indices_1&conf=%22%7B%7D%22" -H "accept: application/json"
    
    # curl -X GET "127.0.0.1:8081/api/experimental/wes/service-info" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/api/experimental/wes/runs" -H "accept: application/json"
    # curl -X POST "127.0.0.1:8081/api/experimental/wes/runs/pracfcizfvmhdefxqdomtxktkbflhgav/cancel" -H "accept: application/json"

    # curl -X GET "127.0.0.1:8081/api/experimental/wes/runs/zlqukumkxxfkumrevclzjcsbyuguhwqy" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/api/experimental/wes/runs/pracfcizfvmhdefxqdomtxktkbflhgav/status" -H "accept: application/json"
    # curl -X POST "127.0.0.1:8081/api/experimental/wes/runs" -H "accept: application/json" -H "Content-Type: multipart/form-data" -F "workflow_attachment[]=@custom-bash.cwl"


    def __init__(self):
        get_dir(DAGS_FOLDER)
        self.include_examples = False
        self.dag_template_with_tmp_folder = "#!/usr/bin/env python3\nfrom cwl_airflow import CWLDAG, CWLJobDispatcher, CWLJobGatherer\ndag = CWLDAG(cwl_workflow='{0}', dag_id='{1}', default_args={{'tmp_folder':'{2}'}})\ndag.create()\ndag.add(CWLJobDispatcher(dag=dag), to='top')\ndag.add(CWLJobGatherer(dag=dag), to='bottom')"
        self.wes_state_conversion = {"running": "RUNNING", "success": "COMPLETE", "failed": "EXECUTOR_ERROR"}
        self.validated_dags = {}  # stores dags' content md5 checksums as keys and one of the statuses ["checking", "success", "error"] as values


    def get_dags(self, dag_ids=[]):
        logging.debug(f"Call get_dags with dag_ids={dag_ids}")
        try:
            dag_ids = dag_ids or self.list_dags()
            logging.debug(f"Processing dags {dag_ids}")
            return {"dags": [{"dag_id": dag_id, "tasks": self.list_tasks(dag_id)} for dag_id in dag_ids]}
        except Exception as err:
            logging.error(f"Failed while running get_dags {err}")
            return {"dags": []}


    def post_dag(self, dag_id=None):
        logging.debug(f"Call post_dag with dag_id={dag_id}")
        try:
            res = self.export_dag(dag_id or ''.join(random.choice(string.ascii_lowercase) for i in range(32)))
            logging.debug(f"Exported DAG {res}")
            return res
        except Exception as err:
            logging.error(f"Failed while running post_dag {err}")
            return connexion.problem(500, "Failed to create dag", str(err))


    def get_dag_runs(self, dag_id=None, run_id=None, execution_date=None, state=None):
        logging.debug(f"Call get_dag_runs with dag_id={dag_id}, run_id={run_id}, execution_date={execution_date}, state={state}")
        try:
            dag_runs = []
            dag_ids = [dag_id] if dag_id else self.list_dags()
            logging.debug(f"Processing dags {dag_ids}")
            for d_id in dag_ids:
                logging.debug(f"Process dag  {d_id}")
                task_ids = self.list_tasks(d_id)
                logging.debug(f"Fetched tasks {task_ids}")
                for dag_run in self.list_dag_runs(d_id, state):
                    logging.debug(f"Process dag run {dag_run['run_id']}, {dag_run['execution_date']}")
                    if run_id and run_id != dag_run["run_id"] or execution_date and execution_date != dag_run["execution_date"]:
                        logging.debug(f"Skip dag_run {dag_run['run_id']}, {dag_run['execution_date']} (run_id or execution_date doesn't match)")
                        continue
                    response_item = {"dag_id": d_id,
                                     "run_id": dag_run["run_id"],
                                     "execution_date": dag_run["execution_date"],
                                     "start_date": dag_run["start_date"],
                                     "state": dag_run["state"],
                                     "tasks": []}
                    logging.debug(f"Get statuses for tasks {task_ids}")
                    for t_id in task_ids:
                        response_item["tasks"].append({"id": t_id, "state": self.task_state(d_id, t_id, dag_run["execution_date"])})
                    dag_runs.append(response_item)
            return {"dag_runs": dag_runs}
        except Exception as err:
            logging.error(f"Failed to call get_dag_runs {err}")
            return {"dag_runs": []}
        

    def post_dag_runs(self, dag_id, run_id=None, conf=None):
        logging.debug(f"Call post_dag_runs with dag_id={dag_id}, run_id={run_id}, conf={conf}")
        try:
            dagrun = self.trigger_dag(dag_id, run_id, conf)
            return {"dag_id": dagrun.dag_id,
                    "run_id": dagrun.run_id,
                    "execution_date": dagrun.execution_date,
                    "start_date": dagrun.start_date,
                    "state": dagrun.state}
        except Exception as err:
            logging.error(f"Failed to call post_dag_runs {err}")
            return connexion.problem(500, "Failed to create dag_run", str(err))


    def post_dag_runs_legacy(self, dag_id):
        data = connexion.request.json
        logging.debug(f"Call post_dag_runs_legacy with dag_id={dag_id}, data={data}")
        return self.post_dag_runs(dag_id, data["run_id"], data["conf"])


    def wait_until_dag_validated(self, dag_path):
        """
        Reads the md5 sum of the DAG python file to see whether it
        was updated. Searches by md5 sum if exactly the same DAG
        python file has been already validated. If it was already
        validated raises exception in case of "error" result
        or does nothing if validation was successfull.
        If validation check for the specific DAG python file is
        still running ("checking" status), sleeps for 1 second and
        checks status again. This approach prevents from running
        multiple processes for exactly the same DAG python file on
        each POST request. Instead of using "airflow list_dags -sd"
        that never ends with exit code other than 0, we use "python3".
        Environment is copied to the subprocess, so it should work
        fine even in portable CWL-Airflow installation
        """
        dag_md5_sum = get_md5_sum(dag_path)
        if dag_md5_sum not in self.validated_dags:
            self.validated_dags[dag_md5_sum] = "checking"
            try:
                check_call(["python3", dag_path], env=environ.copy(), stdout=DEVNULL, stderr=DEVNULL)
                self.validated_dags[dag_md5_sum] = "success"
            except CalledProcessError:
                self.validated_dags[dag_md5_sum] = "error"
        while self.validated_dags[dag_md5_sum] not in ["success", "error"]:
            sleep(1)
        if self.validated_dags[dag_md5_sum] == "error":
            raise ValueError(f"Failed to load DAG from {dag_path}")


    @provide_session
    def create_dag_run(self, dag_id, run_id, conf, session):
        """
        Creates new DagRun. Shouldn't be called with not existent dag_id.
        Raises exception if DagRun for the same DAG with the same run_id
        was previously created
        """
        if session.query(DagRun).filter(DagRun.dag_id==dag_id, DagRun.run_id==run_id).one_or_none():
            raise ValueError(f"dag_run {run_id} for dag_id {dag_id} already exists")
        else:
            run_conf = conf if isinstance(conf, dict) else json.loads(conf)
            dag_run = DagRun(dag_id=dag_id, run_id=run_id, conf=run_conf)
            session.add(dag_run)
            session.commit()
            return dag_run


    @provide_session
    def trigger_dag(self, dag_id, run_id, conf, session):
        """
        Checks if DAG exists (DAG python file was already parsed), then
        creates new DagRun with that DAG.
        If DAG doesn't exist, checks if DAG python file can be loaded,
        waits untill it's validated and only after that creates new DagRun.
        """
        if session.query(DagModel).filter(DagModel.dag_id==dag_id).one_or_none():
            return self.create_dag_run(dag_id, run_id, conf)
        else:
            self.wait_until_dag_validated(path.join(DAGS_FOLDER, dag_id + ".py"))
            return self.create_dag_run(dag_id, run_id, conf)


    def list_dags(self):
        return DagBag(include_examples=self.include_examples).dags.keys()


    def list_tasks(self, dag_id):
        return [t.task_id for t in DagBag(include_examples=self.include_examples).dags[dag_id].tasks]


    def task_state(self, dag_id, task_id, execution_date):
        task_state = TaskInstance(DagBag(include_examples=self.include_examples).dags[dag_id].get_task(task_id=task_id), parsedate(execution_date)).current_state()
        task_state = task_state or "none"
        return task_state


    def list_dag_runs(self, dag_id, state):
        dag_runs = []
        for dag_run in DagRun.find(dag_id=dag_id, state=state):
            dag_runs.append({
                "run_id": dag_run.run_id,
                "state": dag_run.state,
                "execution_date": dag_run.execution_date.isoformat(),
                "start_date": ((dag_run.start_date or '') and dag_run.start_date.isoformat())
            })
        return dag_runs


    def save_attachment(self, attachment, location, exist_ok=False):
        if path.isfile(location) and not exist_ok:
            raise FileExistsError(f"File {location} already exist")
        data = connexion.request.files[attachment]
        data.save(location)
    

    def export_dag(self, dag_id):
        """
        Checks if DAG python file with the same name has been already
        exported. If not, checks if exaclty one of "workflow" and
        "workflow_content" parameters are present in the request. In
        case of "workflow_content" first we need to load a tool from
        it and try to convert it to Workflow (what if it was
        CommandLineTool), then compress it again and write to DAG
        python file. In case of "workflow", first we need to save
        attachment, then try to comvert it to Workflow (the same reason
        as above) and write it to DAG python file.
        """

        dag_path = path.join(DAGS_FOLDER, dag_id + ".py")

        if path.isfile(dag_path):
            raise FileExistsError(f"File {dag_path} already exist")

        if "workflow_content" in (connexion.request.json or []) \
            and "workflow" in connexion.request.files:

            raise ValueError("Only one of the 'workflow' or \
                'workflow_content' parameters can be set")

        if "workflow_content" in (connexion.request.json or []):    # json field might be None, need to take [] as default

            workflow = get_compressed(
                convert_to_workflow(                                # to make sure we are not saving CommandLineTool instead of a Workflow
                    command_line_tool=fast_cwl_load(                # using fast_cwl_load is safe here because we deal with the content of a file
                        connexion.request.json["workflow_content"]
                    )
                )
            )

        elif "workflow" in connexion.request.files:

            workflow = path.join(DAGS_FOLDER, dag_id + ".cwl")
            self.save_attachment("workflow", workflow)
            convert_to_workflow(
                command_line_tool=slow_cwl_load(                    # safer to use slow_cwl_load, because of the possible confusions with all these renaming. TODO: make it less complicate
                    workflow=workflow,
                    only_tool=True
                ),
                location=workflow
            )

        else:

            raise ValueError("At least one of the 'workflow' or \
                'workflow_content' parameters should be set")

        with open(dag_path, "w") as output_stream:
            output_stream.write(DAG_TEMPLATE.format(workflow, dag_id))

        return {"dag_id": dag_id, "cwl_path": workflow, "dag_path": dag_path}


###########################################################################
# WES                                                                     #
###########################################################################
 
    def wes_collect_attachments(self, run_id):
        tempdir = tempfile.mkdtemp(
            dir=get_dir(
                path.abspath(
                    conf_get("cwl", "tmp_folder", path.join(AIRFLOW_HOME, "cwl_tmp_folder"))
                )
            ),
            prefix="run_id_"+run_id+"_"
        )
        logging.debug(f"Save all attached files to {tempdir}")
        for k, ls in iterlists(connexion.request.files):
            logging.debug(f"Process attachment parameter {k}")
            if k == "workflow_attachment":
                for v in ls:
                    try:
                        logging.debug(f"Process attached file {v}")
                        sp = v.filename.split("/")
                        fn = []
                        for p in sp:
                            if p not in ("", ".", ".."):
                                fn.append(secure_filename(p))
                        dest = path.join(tempdir, *fn)
                        if not path.isdir(path.dirname(dest)):
                            get_dir(path.dirname(dest))
                        logging.debug(f"Save {v.filename} to {dest}")
                        v.save(dest)
                    except Exception as err:
                        raise ValueError(f"Failed to process attached file {v}, {err}")
        body = {}       
        for k, ls in iterlists(connexion.request.form):
            logging.debug(f"Process form parameter {k}")
            for v in ls:
                try:
                    if not v:
                        continue
                    if k == "workflow_params":
                        job_file = path.join(tempdir, "job.json")
                        with open(job_file, "w") as f:
                            json.dump(json.loads(v), f, indent=4)
                        logging.debug(f"Save job file to {job_file}")
                        loader = Loader(load.jobloaderctx.copy())
                        job_order_object, _ = loader.resolve_ref(job_file, checklinks=False)
                        body[k] = job_order_object
                    else:
                        body[k] = v
                except Exception as err:
                    raise ValueError(f"Failed to process form parameter {k}, {v}, {err}")

        if "workflow_params" not in body or "workflow_url" not in body:
            raise ValueError("Missing 'workflow_params' or 'workflow_url' in submission")
        
        body["workflow_url"] = path.join(tempdir, secure_filename(body["workflow_url"]))

        return tempdir, body


    def wes_get_service_info(self):
        logging.debug(f"Call wes_get_service_info")
        response = {
            "workflow_type_versions": {
                "CWL": {"workflow_type_version": ["v1.0"]}
            },
            "supported_wes_versions": ["1.0.0"],
            "supported_filesystem_protocols": ["file"],
            "workflow_engine_versions": {
                "cwl-airflow": get_version()
            }
        }
        return response


    def wes_list_runs(self, page_size=None, page_token=None):
        logging.debug(f"Call wes_list_runs with page_size={page_size}, page_token={page_token}")
        logging.debug(f"page_size and page_token are currently ignored by cwl-airflow api server")
        dag_run_info = self.get_dag_runs()
        return [{"run_id": item["run_id"], "state": self.wes_state_conversion[item["state"]]} for item in dag_run_info["dag_runs"] ]


    def wes_run_workflow(self):
        logging.debug(f"Call wes_run_workflow")
        run_id = ''.join(random.choice(string.ascii_lowercase) for i in range(32))
        try:
            tempdir, body = self.wes_collect_attachments(run_id)
            with open(path.join(DAGS_FOLDER, run_id + ".py"), 'x') as o_stream:
                o_stream.write(self.dag_template_with_tmp_folder.format(body["workflow_url"], run_id, tempdir))
            self.post_dag_runs(dag_id=run_id, run_id=run_id, conf=json.dumps({"job": body["workflow_params"]}))
            return {"run_id": run_id}
        except Exception as err:
            logging.debug(f"Failed to run workflow {err}")
            return connexion.problem(500, "Failed to run workflow", str(err))
        

    def wes_get_run_log(self, run_id):
        logging.debug(f"Call wes_get_run_log with {run_id}")
        try:
            dag_run_info = self.get_dag_runs(dag_id=run_id, run_id=run_id)["dag_runs"][0]
            dag_run = DagRun.find(dag_id=run_id, state=None)[0]
            workflow_params = dag_run.conf["job"]
            del workflow_params["id"]
            workflow_outputs = {}
            try:
                results_location = dag_run.get_task_instance(task_id="CWLJobGatherer").xcom_pull()
                with open(results_location, "r") as input_stream:
                    workflow_outputs = json.load(input_stream)
            except Exception as err:
                logging.debug(f"Failed to read workflow results from file. \n {err}")
            return {
                "run_id": run_id,
                "request": {"workflow_params": workflow_params},
                "state": self.wes_state_conversion[dag_run_info["state"]],
                "run_log": {
                    "name": run_id,
                    "cmd": [""],
                    "start_time": dag_run_info["start_date"],
                    "end_time": "",
                    "stdout": "",
                    "stderr": "",
                    "exit_code": ""
                },
                "task_logs": [{"name": task["id"]} for task in dag_run_info["tasks"]],
                "outputs": workflow_outputs
            }
        except Exception as err:
            logging.debug(f"Failed to fetch infromation for {run_id}")
            return {}


    def wes_get_run_status(self, run_id):
        logging.debug(f"Call wes_get_run_status with run_id={run_id}")
        try:
            dag_run_info = self.get_dag_runs(dag_id=run_id, run_id=run_id)["dag_runs"][0]
            return {"run_id": dag_run_info["run_id"], "state": self.wes_state_conversion[dag_run_info["state"]]}
        except Exception as err:
            logging.debug(f"Failed to fetch infromation for {run_id}")
            return {}


    def wes_cancel_run(self, run_id):
        logging.debug(f"Call wes_cancel_run with run_id={run_id}")
        try:
            dag_run = DagRun.find(dag_id=run_id, state=None)[0]
            self.stop_tasks(dag_run)
            self.remove_tmp_data(dag_run)
            return {"run_id": run_id}
        except Exception as err:
            logging.debug(f"Failed to cancel dag run {run_id}, {err}")
            return connexion.problem(500, f"Failed to cancel dag run {run_id}", str(err))


    def stop_tasks(self, dr):
        logging.debug(f"Stop tasks for {dr.dag_id} - {dr.run_id}")
        for ti in dr.get_task_instances():
            logging.debug(f"process {ti.dag_id} - {ti.task_id} - {ti.execution_date} - {ti.pid}")
            if ti.state == State.RUNNING:
                try:
                    process = psutil.Process(ti.pid) if ti.pid else None
                except Exception:
                    logging.debug(f" - cannot find process by PID {ti.pid}")
                    process = None
                ti.set_state(State.FAILED)
                logging.debug(" - set state to FAILED")
                if process:
                    logging.debug(f" - wait for process {ti.pid} to exit")
                    try:
                        cleanup_timeout = int(conf_get("core", "KILLED_TASK_CLEANUP_TIME", 60)) * 2
                        process.wait(timeout=cleanup_timeout)  # raises psutil.TimeoutExpired if timeout. Makes task fail -> DagRun fails
                    except psutil.TimeoutExpired as e:
                        logging.debug(f" - Done waiting for process {ti.pid} to die")


    def remove_tmp_data(self, dr):
        logging.debug(f"Remove tmp data for {dr.dag_id} - {dr.run_id}")
        tmp_folder = None
        try:
            results_location = dr.get_task_instance(task_id="CWLJobDispatcher").xcom_pull()
            with open(results_location, "r") as input_stream:
                dispatcher_outputs = json.load(input_stream)
                tmp_folder = dispatcher_outputs.get("tmp_folder")
        except Exception as err:
            logging.debug(f"Failed to read dispathcer results from file. \n {err}")
        try:
            shutil.rmtree(tmp_folder)
            logging.debug(f"Successfully removed {tmp_folder}")
        except Exception as ex:
            logging.error(f"Failed to delete temporary output directory {tmp_folder}\n {ex}")


#!/usr/bin/env python
import connexion
from connexion.resolver import Resolver

from cwl_airflow.components.api.backend import CWLApiBackend


def run_api_server(args):
    app = connexion.FlaskApp(
        __name__,
        host=args.host,
        port=args.port,
        specification_dir="openapi",
        server="tornado"
    )
    backend = CWLApiBackend()
    app.add_api(
        specification="swagger_configuration.yaml",
        resolver=Resolver(lambda x: getattr(backend, x))
    )
    app.run()


#! /usr/bin/env python3
import json
import jwt
import logging

from airflow.models import Variable
from airflow.utils.state import State
from airflow.hooks.http_hook import HttpHook


CONN_ID = "process_report"
ROUTES = {
    "progress": "airflow/progress",
    "results":  "airflow/results",
    "status":   "airflow/status"
}
PRIVATE_KEY = "process_report_private_key"
ALGORITHM = "process_report_algorithm"

http_hook = HttpHook(method="POST", http_conn_id=CONN_ID)  # won't fail even if CONN_ID doesn't exist


def sign_with_jwt(data):
    try:
        data = jwt.encode(
            payload=data,
            key=Variable.get(PRIVATE_KEY),
            algorithm=Variable.get(ALGORITHM)
        ).decode("utf-8")
    except Exception as err:
        logging.debug(f"Failed to sign data with JWT key. \n {err}")
    return data


def post_progress(context, from_task=None):
    from_task = False if from_task is None else from_task
    try:
        dag_run = context["dag_run"]
        len_tis = len(dag_run.get_task_instances())
        len_tis_success = len(dag_run.get_task_instances(state=State.SUCCESS)) + int(from_task)
        data = sign_with_jwt(
            {
                "state": dag_run.state,
                "dag_id": dag_run.dag_id,
                "run_id": dag_run.run_id,
                "progress": int(len_tis_success / len_tis * 100),
                "error": context["reason"] if dag_run.state == State.FAILED else ""
            }
        )
        http_hook.run(endpoint=ROUTES["progress"], json={"payload": data})
    except Exception as err:
        logging.debug(f"Failed to POST progress updates. \n {err}")


def post_results(context):
    """
    Results are collected from the task with id "CWLJobGatherer". We cannot use
    isinstance(task, CWLJobGatherer) to find the proper task because of the
    endless import loop (file where we define CWLJobGatherer class import this
    file). If CWLDAG is contsructed with custom gatherer node, posting results
    might not work. We need to except missing results file as the same callback
    is used for clean_dag_run DAG
    """
    try:
        dag_run = context["dag_run"]
        results = {}
        try:
            results_location = context["ti"].xcom_pull(task_ids="CWLJobGatherer")
            with open(results_location, "r") as input_stream:
                results = json.load(input_stream)
        except Exception as err:
            logging.debug(f"Failed to read results. \n {err}")
        data = sign_with_jwt(
            {
                "dag_id": dag_run.dag_id,
                "run_id": dag_run.run_id,
                "results": results
            }
        )
        http_hook.run(endpoint=ROUTES["results"], json={"payload": data})
    except Exception as err:
        logging.debug(f"Failed to POST results. \n {err}")


def post_status(context):
    try:
        dag_run = context["dag_run"]
        ti = context["ti"]
        data = sign_with_jwt(
            {
                "state": ti.state,
                "dag_id": dag_run.dag_id,
                "run_id": dag_run.run_id,
                "task_id": ti.task_id
            }
        )
        http_hook.run(endpoint=ROUTES["status"], json={"payload": data})
    except Exception as err:
        logging.debug(f"Failed to POST status updates. \n {err}")


def task_on_success(context):
    post_progress(context, True)
    post_status(context)


def task_on_failure(context):
    post_status(context)


def task_on_retry(context):
    post_status(context)


def dag_on_success(context):
    post_progress(context)
    post_results(context)


def dag_on_failure(context):
    post_progress(context)



#! /usr/bin/env python3
import os
from tempfile import mkdtemp

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from cwl_airflow.utilities.helpers import (
    get_dir,
    dump_json,
    get_absolute_path
)
from cwl_airflow.utilities.cwl import (
    load_job,
    get_temp_folders
)
from cwl_airflow.utilities.report import post_status
from cwl_airflow.utilities.loggers import setup_cwl_logger


class CWLJobDispatcher(BaseOperator):


    @apply_defaults  # in case someone decided to overwrite default_args from the DAG
    def __init__(
        self,
        task_id,
        *args, **kwargs
    ):
        super().__init__(task_id=task_id, *args, **kwargs)


    def execute(self, context):
        """
        Loads job Object from the context. Sets "tmp_folder" and "output_folder"
        if they have not been set before in the job. In case "tmp_folder" and/or
        "output_folder" were read from the job and are relative, resolves paths
        relative to the "tmp_folder" and/or "outputs_folder" from "cwl_args".
        Dumps step outputs as a json file into "tmp_folder". Writes to X-Com report
        file location.
        """

        setup_cwl_logger(context["ti"])
        post_status(context)

        # for easy access
        dag_id = context["dag"].dag_id
        workflow = context["dag"].workflow
        run_id = context["run_id"].replace(":", "_").replace("+", "_")  # to make it dumpable by json
        cwl_args = context["dag"].default_args["cwl"]

        # Loads job from dag_run configuration. Sets defaults from "workflow". Fails on missing input files
        job_data = load_job(
            workflow=workflow,
            job=context["dag_run"].conf["job"],
            cwl_args=cwl_args
        )

        job_data["tmp_folder"] = get_dir(
            get_absolute_path(
                job_data.get(
                    "tmp_folder",
                    mkdtemp(
                        dir=cwl_args["tmp_folder"],
                        prefix=dag_id+"_"+run_id+"_"
                    )
                ),
                cwl_args["tmp_folder"]
            )
        )

        job_data["outputs_folder"] = get_dir(
            get_absolute_path(
                job_data.get(
                    "outputs_folder",
                    os.path.join(
                        cwl_args["outputs_folder"],
                        dag_id,
                        run_id
                    )
                ),
                cwl_args["outputs_folder"]
            )
        )

        _, _, _, step_report = get_temp_folders(
            task_id=self.task_id,
            job_data=job_data
        )

        dump_json(job_data, step_report)

        return step_report



#! /usr/bin/env python3
import os

from copy import deepcopy
from airflow.models import DAG
from airflow.utils.dates import days_ago

from cwl_airflow.utilities.cwl import (
    fast_cwl_load,
    get_items,
    get_default_cwl_args
)
from cwl_airflow.extensions.operators.cwlstepoperator import CWLStepOperator
from cwl_airflow.extensions.operators.cwljobdispatcher import CWLJobDispatcher
from cwl_airflow.extensions.operators.cwljobgatherer import CWLJobGatherer
from cwl_airflow.utilities.report import (
    dag_on_success,
    dag_on_failure,
    task_on_success,
    task_on_failure,
    task_on_retry
)


class CWLDAG(DAG):

    def __init__(
        self,
        dag_id,          # the id of the DAG
        workflow,        # absolute path to the CWL workflow file or utf-8 string to include base64 encoded zlib compressed utf-8 workflow file content
        dispatcher=None, # custom job dispatcher. Will be assigned automatically to the same DAG. Default CWLJobDispatcher
        gatherer=None,   # custom job gatherer. Will be assigned automatically to the same DAG. Default CWLJobGatherer
        *args, **kwargs  # see DAG class for additional parameters
    ):
        """
        Updates kwargs with the required defaults if they were not explicitely provided
        by user. dispatcher and gatherer are set to CWLJobDispatcher() and CWLJobGatherer()
        if those were not provided by user. If user sets his own operators for dispatcher
        and gatherer, "default_args" will not be inherited. User needs to set up proper
        agruments by himself. Also, dag results will not be posted from the custom dispatcher.
        """

        self.workflow = workflow
        self.__setup_params(kwargs)

        super().__init__(dag_id=dag_id, *args, **kwargs)

        self.workflow_tool = fast_cwl_load(         # keeps only the tool (CommentedMap object)
            workflow=self.workflow,
            cwl_args=kwargs["default_args"]["cwl"]  # in case user has overwritten some of the default parameters
        )

        self.dispatcher = CWLJobDispatcher(
            dag=self,                               # need dag=self otherwise new operator will not get proper default_args
            task_id="CWLJobDispatcher"
        ) if dispatcher is None else dispatcher

        self.gatherer = CWLJobGatherer(
            dag=self,                               # need dag=self otherwise new operator will not get proper default_args
            task_id="CWLJobGatherer"
        ) if gatherer is None else gatherer

        self.__assemble()


    def __setup_params(self, kwargs):
        """
        Updates kwargs with default values if those were not
        explicitely set on CWLDAG creation. "start_date" is set
        to days_ago(180) assuming that DAG run is not supposed
        to be queued longer then half a year:)
        """

        # default args provided by user. Use deepcopy to prevent from changing in place
        user_default_args = deepcopy(kwargs.get("default_args", {}))
        
        # get all the parameters required by cwltool with preset by user defaults
        required_cwl_args = get_default_cwl_args(
            preset_cwl_args=user_default_args.get("cwl", {})
        )

        # update default args provided by user with required by cwltool args
        user_default_args.update({
            "cwl": required_cwl_args
        })

        # default arguments required by CWL-Airflow (no need to put it in a separate function so far)
        required_default_args = {
            "start_date": days_ago(180),
            "email_on_failure": False,
            "email_on_retry": False,
            "on_failure_callback": task_on_failure,
            "on_success_callback": task_on_success,
            "on_retry_callback": task_on_retry
        }
        
        # Updated default arguments required by CWL-Airflow with those that are provided by user for cwltool
        required_default_args.update(user_default_args)

        # update kwargs with correct default_args and callbacks if those were not set by user
        kwargs.update(
            {
                "default_args": required_default_args,
                "on_failure_callback": kwargs.get("on_failure_callback", dag_on_failure),
                "on_success_callback": kwargs.get("on_success_callback", dag_on_success),
                "schedule_interval": None
            }
        )


    def __assemble(self):
        """
        Creates DAG based on the parsed CWL workflow structure.
        Assignes dispatcher and gatherer tasks
        """

        # TODO: add support for CommandLineTool and ExpressionTool
        # TODO: add colors for Tasks?

        task_by_id = {}         # to get airflow task assosiated with workflow step by its id
        task_by_out_id = {}     # to get airflow task assosiated with workflow step by its out id
        
        for step_id, step_data in get_items(self.workflow_tool["steps"]):
            task_by_id[step_id] = CWLStepOperator(dag=self, task_id=step_id)
            for step_out_id, _ in get_items(step_data["out"]):
                task_by_out_id[step_out_id] = task_by_id[step_id]

        for step_id, step_data in get_items(self.workflow_tool["steps"]):
            for step_in_id, step_in_data in get_items(step_data.get("in", [])):           # step might not have "in"
                for step_in_source, _ in get_items(step_in_data.get("source", [])):       # "in" might not have "source"
                    try:
                        task_by_id[step_id].set_upstream(task_by_out_id[step_in_source])  # connected to another step
                    except KeyError:
                        task_by_id[step_id].set_upstream(self.dispatcher)                 # connected to dispatcher
            if not step_data.get("in", []):                                               # safety measure in case "in" was empty
                task_by_id[step_id].set_upstream(self.dispatcher)                         # connected to dispatcher


        for _, output_data in get_items(self.workflow_tool["outputs"]):
            for output_source_id, _ in get_items(output_data["outputSource"]):            # in case "outputSource" is a list
                try:
                    self.gatherer.set_upstream(task_by_out_id[output_source_id])          # connected to another step
                except KeyError:
                    self.gatherer.set_upstream(self.dispatcher)                           # connected to dispatcher

        # safety measure in case of very specific workflows
        # if gatherer happened to be not connected to anything, connect it to all "leaves"
        # if dispatcher happened to be not connected to anything, connect it to all "roots"

        if not self.gatherer.upstream_list:
            self.gatherer.set_upstream([task for task in task_by_id.values() if not task.downstream_list])

        if not self.dispatcher.downstream_list:
            self.dispatcher.set_downstream([task for task in task_by_id.values() if not task.upstream_list])



import argparse

from os import environ, getcwd
from tempfile import mkdtemp

from cwl_airflow.utilities.helpers import (
    get_version,
    get_absolute_path,
    CleanAirflowImport,
    load_yaml
)

from cwl_airflow.components.api.server import run_api_server
from cwl_airflow.components.init.config import run_init_config
from cwl_airflow.components.test.conformance import run_test_conformance

with CleanAirflowImport():
    from airflow.configuration import (
        get_airflow_home,
        get_airflow_config
    )


def get_normalized_args(args, skip_list=None, cwd=None):
    """
    Converts all relative path arguments to absolute
    ones relatively to the cwd or current working directory.
    Skipped arguments and None will be returned unchanged.
    """

    cwd = getcwd() if cwd is None else cwd
    skip_list = [] if skip_list is None else skip_list

    normalized_args = {}
    for key, value in args.__dict__.items():
        if key not in skip_list and value is not None:
            if isinstance(value, list):
                for v in value:
                    normalized_args.setdefault(key, []).append(
                        get_absolute_path(v, cwd)
                    )
            else:
                normalized_args[key] = get_absolute_path(value, cwd)
        else:
            normalized_args[key] = value
    return argparse.Namespace(**normalized_args)


def get_parser():
    """
    Defines arguments for parser. Inlcudes two subparsers for
    api server and init components.
    """

    parent_parser = argparse.ArgumentParser(add_help=False)
    general_parser = argparse.ArgumentParser(
        description="CWL-Airflow: a lightweight pipeline manager \
            supporting Common Workflow Language"
    )
    subparsers = general_parser.add_subparsers()
    subparsers.required = True

    general_parser.add_argument(                       
        "--version",
        action="version",
        version=get_version(),
        help="Print current version and exit"
    )

    # API
    api_parser = subparsers.add_parser(
        "api",
        parents=[parent_parser],
        help="Run API server"
    )
    api_parser.set_defaults(func=run_api_server)
    api_parser.add_argument(
        "--port", 
        type=int,
        default=8081,
        help="Set port to run API server. Default: 8081",
    )
    api_parser.add_argument(
        "--host", 
        type=str,
        default="127.0.0.1",
        help="Set host to run API server. Default: 127.0.0.1"
    )

    # Init
    init_parser = subparsers.add_parser(
        "init",
        parents=[parent_parser],
        help="Run initial configuration"
    )
    init_parser.set_defaults(func=run_init_config)
    init_parser.add_argument(
        "--home", 
        type=str,
        default=get_airflow_home(),
        help="Set path to Airflow home directory. \
            Default: first try AIRFLOW_HOME then '~/airflow'"
    )
    init_parser.add_argument(
        "--config", 
        type=str,
        help="Set path to Airflow configuration file. \
            Default: first try AIRFLOW_CONFIG then '[airflow home]/airflow.cfg'"
    )
    init_parser.add_argument(
        "--upgrade",
        action="store_true",
        help="Upgrade old CWLDAG files to the latest format. \
            Default: False"
    )

    # Test (for running conformance tests)
    test_parser = subparsers.add_parser(
        "test",
        parents=[parent_parser],
        help="Run conformance tests"
    )
    test_parser.set_defaults(func=run_test_conformance)
    test_parser.add_argument(
        "--suite", 
        type=str,
        required=True,
        help="Set path to the conformance test suite"
    )
    test_parser.add_argument(
        "--tmp",
        type=str,
        default=mkdtemp(),
        help="Set path to the temp folder. Default: /tmp"
    )
    test_parser.add_argument(
        "--port", 
        type=int,
        default=3070,
        help="Set port to listen for DAG results and status updates. \
            Should correspond to the port from 'process_report' connection. \
            Default: 3070"
    )
    test_parser.add_argument(
        "--api",
        type=str,
        default="http://127.0.0.1:8081",
        help="Set CWL-Airflow API's base url (IP address and port). \
            Default: http://127.0.0.1:8081"
    )
    test_parser.add_argument(
        "--range", 
        type=str,
        help="Set test range to run, format 1,3-6,9. \
            Order corresponds to the tests from --suite file, starting from 1. \
            Default: run all tests"
    )
    test_parser.add_argument(
        "--spin",
        action="store_true",
        help="Display spinner wher running. Useful for CI that tracks activity. \
            Default: do not spin"
    )
    test_parser.add_argument(
        "--embed",
        action="store_true",
        help="Embed base64 encoded zlib compressed workflow content into the DAG python file. \
            Default: False"
    )

    return general_parser


def assert_and_fix_args_for_init(args):
    """
    Asserts, fixes and sets parameters from init parser.
    """
    
    if args.config is None:
        args.config = get_airflow_config(args.home)


def assert_and_fix_args_for_test(args):
    """
    Asserts, fixes and sets parameters from test parser.

    Tries to convert --range argument to a list of indices.
    If --range wasn't set or indices are not valid, set it
    to include all tests from --suite. Dublicates are removed.
    This function should never fail.
    """

    suite_data = load_yaml(args.suite)
    suite_len = len(suite_data)
    try:
        selected_indices = []
        for interval in args.range.split(","):
            parsed_interval = [int(i) - 1 for i in interval.split("-")]   # switch to 0-based coodrinates
            if len(parsed_interval) == 2:
                # safety check
                if parsed_interval[0] >= parsed_interval[1]: raise ValueError
                if parsed_interval[0] >= suite_len: raise ValueError
                if parsed_interval[1] >= suite_len: raise ValueError
                selected_indices.extend(
                    range(parsed_interval[0], parsed_interval[1] + 1)  # need closed interval
                )
            elif len(parsed_interval) == 1:
                # safety check
                if parsed_interval[0] >= suite_len: raise ValueError
                selected_indices.append(parsed_interval[0])
            else:
                raise ValueError
    except (AttributeError, IndexError, ValueError):
        selected_indices = list(range(0, suite_len))

    args.range = sorted(set(selected_indices))           # convert to set to remove possible dublicates


def assert_and_fix_args(args):
    """
    Should be used to assert and fix parameters.
    Also can be used to set default values for not
    set parameters in case the later ones depend on other
    parameters that should be first parsed by argparser
    """

    if args.func == run_init_config:
        assert_and_fix_args_for_init(args)
    elif args.func == run_test_conformance:
        assert_and_fix_args_for_test(args)
    else:
        pass  # TODO: once needed, put here something like assert_and_fix_args_for_api


def parse_arguments(argsl, cwd=None):
    """
    Parses provided through argsl arguments.
    All relative paths will be resolved relative
    to the cwd or current working directory.
    Set argsl to [""] if empty to prevent argparse from failing
    """

    cwd = getcwd() if cwd is None else cwd
    argsl = argsl + [""] if len(argsl) == 0 else argsl
    
    args, _ = get_parser().parse_known_args(argsl)
    args = get_normalized_args(
        args,
        ["func", "port", "host", "api", "range", "spin", "embed", "upgrade"],
        cwd
    )
    assert_and_fix_args(args)
    return args


import logging
from cwl_airflow.utilities.cwl import conf_get


def setup_cwl_logger(ti, level=None):
    """
    Sets logging level of cwltool logger to correspond LOGGING_LEVEL
    from airflow.cfg. Configures handler based on the task instance
    to redirect output to the proper file. Suppresses those loggers
    from cwltool or related packages that spam.
    Note: maybe we will need to remove StreamHandler <stderr> handler
    from cwltool logger in case we see undesired outputs in the airflow
    logs but not in the separate files.
    """

    level = conf_get("core", "LOGGING_LEVEL", "INFO").upper() if level is None else level
    cwl_logger = logging.getLogger("cwltool")
    for handler in cwl_logger.handlers:
        try:
            handler.set_context(ti)
        except AttributeError:
            pass
    cwl_logger.setLevel(level)

    loggers_to_suppress = ["rdflib.term", "salad", "requests", "urllib3"]
    for logger_name in loggers_to_suppress:
        logger = logging.getLogger(logger_name)
        logger.setLevel("ERROR")


import sys
import os
import re
import uuid
import shutil
import logging

from subprocess import run, DEVNULL, CalledProcessError

from cwl_airflow.utilities.helpers import (
    get_dir,
    CleanAirflowImport
)

with CleanAirflowImport():
    from airflow import models
    from airflow.configuration import conf
    from airflow.utils.db import merge_conn
    from airflow.utils.dag_processing import list_py_file_paths
    from cwl_airflow.utilities.cwl import overwrite_deprecated_dag


def run_init_config(args):
    """
    Runs sequence of steps required to configure CWL-Airflow
    for the first time. Safe to run several times
    """

    init_airflow_db(args)
    patch_airflow_config(args.config)
    # add_connections(args)
    if args.upgrade:
        upgrade_dags(args.config)
    copy_dags(args.home)


def init_airflow_db(args):
    """
    Sets AIRFLOW_HOME and AIRFLOW_CONFIG from args.
    Call airflow initdb from subprocess to make sure
    that the only two things we should care about
    are AIRFLOW_HOME and AIRFLOW_CONFIG
    """

    custom_env = os.environ.copy()
    custom_env["AIRFLOW_HOME"] = args.home
    custom_env["AIRFLOW_CONFIG"] = args.config
    try:
        run(
            ["airflow", "initdb"],  # TODO: check what's the difference initdb from updatedb
            env=custom_env,
            check=True,
            stdout=DEVNULL,
            stderr=DEVNULL
        )
    except (CalledProcessError, FileNotFoundError) as err:
        logging.error(f"""Failed to run 'airflow initdb'. Exiting.\n{err}""")
        sys.exit(1)


def patch_airflow_config(airflow_config):
    """
    Updates provided Airflow configuration file to include defaults for cwl-airflow.
    If something went wrong, restores the original airflow.cfg from the backed up copy
    """

    # TODO: add cwl section with the following parameters:
    # - singularity
    # - use_container

    patches = [
        ["sed", "-i", "-e", "s/^dags_are_paused_at_creation.*/dags_are_paused_at_creation = False/g", airflow_config],
        ["sed", "-i", "-e", "s/^load_examples.*/load_examples = False/g", airflow_config],
        ["sed", "-i", "-e", "s/^logging_config_class.*/logging_config_class = cwl_airflow.config_templates.airflow_local_settings.DEFAULT_LOGGING_CONFIG/g", airflow_config],
        ["sed", "-i", "-e", "s/^hide_paused_dags_by_default.*/hide_paused_dags_by_default = True/g", airflow_config]
    ]

    airflow_config_backup = airflow_config + "_backup_" + str(uuid.uuid4())
    try:
        shutil.copyfile(airflow_config, airflow_config_backup)
        for patch in patches:
            run(
                patch,
                shell=False,  # for proper handling of filenames with spaces
                check=True,
                stdout=DEVNULL,
                stderr=DEVNULL
            )
    except (CalledProcessError, FileNotFoundError) as err:
        logging.error(f"""Failed to patch Airflow configuration file. Restoring from the backup and exiting.\n{err}""")
        if os.path.isfile(airflow_config_backup):
            shutil.copyfile(airflow_config_backup, airflow_config)
        sys.exit(1)
    finally:
        if os.path.isfile(airflow_config_backup):
            os.remove(airflow_config_backup)


def upgrade_dags(airflow_config):
    """
    Corrects old style DAG python files into the new format.
    Reads configuration from "airflow_config". Uses standard
    "conf.get" instead of "conf_get", because the fields we
    use are always set. Copies all deprecated dags into the 
    "deprecated_dags" folder, adds deprecated DAGs to the
    ".airflowignore" file created within that folder. Original
    DAG files are replaced with the new ones (with base64
    encoded zlib compressed workflow content), original workflow
    files remain unchanged.
    """

    conf.read(airflow_config)
    dags_folder = conf.get("core", "dags_folder")
    for dag_location in list_py_file_paths(                     # will skip all DAGs from ".airflowignore"
        directory=dags_folder,
        safe_mode=conf.get("core", "dag_discovery_safe_mode"),  # use what user set in his config
        include_examples=False
    ):
        overwrite_deprecated_dag(                               # upgrades only deprecated DAGs, skips others
            dag_location=dag_location,
            deprecated_dags_folder=os.path.join(
                dags_folder,
                "deprecated_dags"
            )
        )


def copy_dags(airflow_home, source_folder=None):
    """
    Copies *.py files (dags) from source_folder (default ../../extensions/dags)
    to dags_folder, which is always {airflow_home}/dags
    """

    if source_folder is None:
        source_folder = os.path.join(
            os.path.dirname(
                os.path.abspath(
                    os.path.join(__file__, "../../")     
                )
            ), 
            "extensions/dags",
        )

    target_folder = get_dir(os.path.join(airflow_home, "dags"))
    for root, dirs, files in os.walk(source_folder):
        for filename in files:
            if re.match(".*\\.py$", filename) and filename != "__init__.py":
                if not os.path.isfile(os.path.join(target_folder, filename)):
                    shutil.copy(os.path.join(root, filename), target_folder)

# not used anymore
def add_connections(args):
    """
    Sets AIRFLOW_HOME and AIRFLOW_CONFIG from args.
    Call 'airflow connections --add' from subproces to make sure that
    the only two things we should care about are AIRFLOW_HOME and
    AIRFLOW_CONFIG. Adds "process_report" connections to the Airflow DB
    that is used to report workflow execution progress and results.
    """

    custom_env = os.environ.copy()
    custom_env["AIRFLOW_HOME"] = args.home
    custom_env["AIRFLOW_CONFIG"] = args.config
    try:
        run(
            [
                "airflow", "connections", "--add",
                "--conn_id", "process_report",
                "--conn_type", "http",
                "--conn_host", "localhost",
                "--conn_port", "3070",
                "--conn_extra", "{\"endpoint\":\"/airflow/\"}"
            ],
            env=custom_env,
            check=True,
            stdout=DEVNULL,
            stderr=DEVNULL
        )
    except (CalledProcessError, FileNotFoundError) as err:
        logging.error(f"""Failed to run 'airflow connections --add'. Exiting.\n{err}""")
        sys.exit(1)


#! /usr/bin/env python3
import os
import subprocess
import time
from setuptools import setup, find_packages


GIT_VERSION_FILE = os.path.join("cwl_airflow","git_version")
HERE = os.path.abspath(os.path.dirname(__file__))


def get_git_tag():
    return subprocess.check_output(["git", "describe", "--contains"], text=True).split("^")[0].strip()


def get_git_timestamp():
    gitinfo = subprocess.check_output(
        ["git", "log", "--first-parent", "--max-count=1", "--format=format:%ct", "."], text=True).strip()
    return time.strftime("%Y%m%d%H%M%S", time.gmtime(int(gitinfo)))


def get_description():
    README = os.path.join(HERE, "README.md")
    with open(README, "r") as f:
        return f.read()


def get_version():
    """
    Tries to get pachage version with following order:
    0. default version
    1. from git_version file - when installing from pip, this is the only source to get version
    2. from tag
    3. from commit timestamp
    Updates/creates git_version file with the package version
    :return: package version
    """
    version = "1.2.0"                                      # set default version
    try:
        with open(GIT_VERSION_FILE, "r") as input_stream:  # try to get version info from file
            version = input_stream.read()
    except Exception:
        pass
    try:
        version = get_git_tag()                            # try to get version info from the closest tag
    except Exception:
        try:
            version = "1.2." + get_git_timestamp()         # try to get version info from commit date
        except Exception:
            pass
    try:
        with open(GIT_VERSION_FILE, "w") as output_stream: # save updated version to file (or the same)
            output_stream.write(version)
    except Exception:
        pass
    return version


EXTRAS_REQUIRE = {
    "celery": [
        "celery~=4.3",
        "flower>=0.7.3,<1.0",
        "kombu==4.6.3;python_version<'3.0'",
        "tornado>=4.2.0,<6.0"
    ],
    "mysql": [
        "mysqlclient>=1.3.6,<1.4"
    ],
    "statsd": [
        "statsd>=3.3.0,<4.0"
    ],
    "rabbitmq": [
        "amqp"
    ],
    "crypto": [
        "cryptography>=0.9.3"
    ]
}


setup(
    name="cwl-airflow",
    description="Python package to extend Airflow functionality with CWL v1.1 support",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    version=get_version(),
    url="https://github.com/Barski-lab/cwl-airflow",
    download_url="https://github.com/Barski-lab/cwl-airflow",
    author="Michael Kotliar",
    author_email="misha.kotliar@gmail.com",
    license="Apache-2.0",
    include_package_data=True,
    packages=find_packages(
        exclude=["docs", "tests", "dev"]
    ),
    extras_require=EXTRAS_REQUIRE,
    install_requires=[
        "apache-airflow==1.10.12",
        "cwltool==3.0.20200710214758",
        "cwltest==2.0.20200626112502",
        "jsonmerge",
        "pyjwt",
        "connexion",
        "tornado",
        "docker",
        "swagger-ui-bundle"
    ],
    zip_safe=False,
    scripts=["cwl_airflow/bin/cwl-airflow"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Medical Science Apps."
    ]
)



#! /usr/bin/env python3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from cwl_airflow.utilities.cwl import (
    relocate_outputs,
    collect_reports
)
from cwl_airflow.utilities.report import post_status
from cwl_airflow.utilities.loggers import setup_cwl_logger


class CWLJobGatherer(BaseOperator):

    @apply_defaults  # in case someone decided to overwrite default_args from the DAG
    def __init__(
        self,
        task_id,
        *args, **kwargs
    ):
        super().__init__(task_id=task_id, *args, **kwargs)
        

    def execute(self, context):
        """
        Loads and merges data from report files of all finished tasks in a DAG.
        Relocates results to the "outputs_folder", removes "tmp_folder"
        """

        setup_cwl_logger(context["ti"])
        post_status(context)

        _, workflow_report = relocate_outputs(
            workflow=context["dag"].workflow,
            job_data=collect_reports(context),
            cwl_args=context["dag"].default_args["cwl"],
        )
     
        return workflow_report
